// file: $isip_ifc/class/algo/DynamicFilter/DynamicFilter.h
//

// make sure definitions are only made once
//
#ifndef ISIP_DYNAMIC_FILTER
#define ISIP_DYNAMIC_FILTER

// isip include files
//
#ifndef ISIP_ALGORITHM_BASE
#include <AlgorithmBase.h>
#endif

#ifndef ISIP_PAIR
#include <Pair.h>
#endif

// DynamicFilter: This class iteratively filters the reconstructed phase space in
// in order to remove distortions caused by noise. Filtering is achieved by applying
// a smoothing transformation to each point, Y(k), of the attractor. The transformation
// for each point is computed using m of its nearest neighbors which are collected to
// form a matrix G(k). The structure of this neighborhood can be represented by its set
// of eigenvectors which we estimate using SVD of G(k):
//
//                       G(k) = U*S*V'
//
// where U is an m by m matrix formed by the eigenvectors of the G(k)*G(k)' matrix, S 
// contains the singular values, and V is a d by d matrix (where d is the number of 
// dimensions of the phase space) formed by the eigenvectors of the covariance matrix 
// G(k)'G(k).
//
// It is assumed that the smaller eigenvalues correspond to noise, so in order to suppress
// these, each point is projected onto the subspace spanned by its larger principal 
// components. This transformation is applied to each point on the attractor for a 
// user-specified number of iterations. For each iteration, the point is contracted by a
// factor between alpha_min and alpha_max (usually, 0.1-1.0). This factor is linearly
// increased for each iteration.
//
// For more information, see:
// 
// V. Pitsikalis, P. Maragos,  "Filtered Dynamics and Fractal Dimensions for 
// Noisy Speech Recognition," IEEE Signal Processing Letters, vol. 13, no. 11, 
// pp. 711-715, November 2006. 
//
class DynamicFilter : public AlgorithmBase {

  //---------------------------------------------------------------------------
  //
  // public constants
  //
  //---------------------------------------------------------------------------
public:

  // define the class name
  //
  static const String CLASS_NAME;

  // define the signal type
  //
  enum ALGORITHM { NORMAL = 0, SVD_CLEAN, DEF_ALGORITHM = NORMAL };

  // define the implementation type (this is a dummy implementation)
  //
  enum IMPLEMENTATION { REGULAR = 0, DEF_IMPLEMENTATION = REGULAR };

  // define static NameMap objects
  //
  static const NameMap ALGO_MAP;
  static const NameMap IMPL_MAP;
  
  //----------------------------------------
  //
  // i/o related constants
  //
  //----------------------------------------  

  static const String DEF_PARAM;
  static const String PARAM_ALGORITHM;
  static const String PARAM_IMPLEMENTATION;
  static const String PARAM_NUM_NEIGHBORS;
  static const String PARAM_ALPHA_MIN;
  static const String PARAM_ALPHA_MAX;
  static const String PARAM_NUM_ITERATIONS;
  static const String PARAM_NUM_EIGENS;

  //----------------------------------------
  //
  // default values and arguments
  //
  //----------------------------------------  

  // define the default value(s) of the class data
  //
  static const int32 DEF_NUM_NEIGHBORS = 10;
  static const float32 DEF_ALPHA_MIN = 0.1;
  static const float32 DEF_ALPHA_MAX = 10;
  static const int32 DEF_NUM_ITERATIONS = 10;
  static const int32 DEF_NUM_EIGENS = 1;

  // define the default value for infinity
  //
  static const float32 DYNAMIC_FILTER_INF = 1.0e16;

  // define default argument(s)
  //
  static const AlgorithmData::COEF_TYPE DEF_COEF_TYPE = AlgorithmData::RPS; 
  
  //----------------------------------------
  //
  // error codes 
  //
  //----------------------------------------  

  static const int32 ERR = 73600;

  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:

  // algorithm name
  //
  ALGORITHM algorithm_d;

  // dummy implementation name
  // there is no implementation method in the DynamicFilter class,
  // this is provided to maintain the consistency
  // across all the algorithm classes
  //
  IMPLEMENTATION implementation_d;

  // number of filter iterations
  //
  Long num_iterations_d;

  // range of transformation factor. for each iteration i, the point
  // in the attactor will be moved by a factor of
  //
  //         i * (alpha_max_d - alpha_min_d)/num_iterations_d + alpha_min_d
  //
  //
  Float alpha_min_d;
  Float alpha_max_d;

  // number of nearest neighbors to use to compute transformation
  //
  Long num_neighbors_d;

  // number of eigen vectors of neighborhood matrix to use for projection
  //
  Long num_eigens_d;

  // parameters related to the algorithm specification
  //
  
  // a static memory manager
  //
  static MemoryManager mgr_d;
  
  //---------------------------------------------------------------------------
  //
  // required public methods
  //
  //---------------------------------------------------------------------------
public:
    
  // method: name
  //
  static const String& name() {
    return CLASS_NAME;
  }

  // other static methods
  //
  static bool8 diagnose(Integral::DEBUG debug_level);
  
  // debug methods:
  //  setDebug is inherited from the AlgorithmBase class
  //
  bool8 debug(const unichar* msg) const;
  
  // method: destructor
  //
  ~DynamicFilter() {}

  // method: default constructor
  //
  DynamicFilter(ALGORITHM algorithm = DEF_ALGORITHM,
		IMPLEMENTATION implementation = DEF_IMPLEMENTATION,
		int32 num_iterations = DEF_NUM_ITERATIONS,
		float32 alpha_min = DEF_ALPHA_MIN,
		float32 alpha_max = DEF_ALPHA_MAX,
		int32 num_neighbors = DEF_NUM_NEIGHBORS,
		int32 num_eigens = DEF_NUM_EIGENS) {

    // initialize protected data
    //    
    algorithm_d = algorithm;
    implementation_d = implementation;
    num_iterations_d = num_iterations;
    alpha_min_d = alpha_min;
    alpha_max_d = alpha_max;
    num_neighbors_d = num_neighbors;
    num_eigens_d = num_eigens;

  }

  // assign methods:
  //
  bool8 assign(const DynamicFilter& arg) {
    algorithm_d = arg.algorithm_d;
    implementation_d = arg.implementation_d;
    num_neighbors_d = arg.num_neighbors_d;
    alpha_min_d = arg.alpha_min_d;
    alpha_max_d = arg.alpha_max_d;
    num_iterations_d = arg.num_iterations_d;
    num_eigens_d = arg.num_eigens_d;
    return AlgorithmBase::assign(arg);
  }

  // method: copy constructor
  //
  DynamicFilter(const DynamicFilter& arg) {
    assign(arg);
  }
  
  // method: operator=
  //
  DynamicFilter& operator= (const DynamicFilter& arg) {
    assign(arg);
    return *this;
  }
  
  // i/o methods:
  //
  int32 sofSize() const;
  
  bool8 read(Sof& sof, int32 tag, const String& name = CLASS_NAME);

  bool8 write(Sof& sof, int32 tag, const String& name = CLASS_NAME) const;
  
  bool8 readData(Sof& sof, const String& pname = DEF_PARAM,
                   int32 size = SofParser::FULL_OBJECT,
                   bool8 param = true,
                   bool8 nested = false);
  
  bool8 writeData(Sof& sof, const String& pname = DEF_PARAM) const;

  // method: eq
  //
  bool8 eq(const DynamicFilter& arg) const {
    return ((algorithm_d == arg.algorithm_d) &&
	    (implementation_d == arg.implementation_d) &&
	    (num_neighbors_d == arg.num_neighbors_d) &&
	    (alpha_min_d == arg.alpha_min_d) &&
	    (alpha_max_d == arg.alpha_max_d) &&
	    (num_iterations_d == arg.num_iterations_d) &&
	    (num_eigens_d == arg.num_eigens_d));
  }
  
  // method: new
  //
  static void* operator new(size_t size) {
    return mgr_d.get();
  }

  // method: new[]
  //
  static void* operator new[](size_t size) {
    return mgr_d.getBlock(size);
  }

  // method: delete
  //
  static void operator delete(void* ptr) {
    mgr_d.release(ptr);
  }

  // method: delete[]
  //
  static void operator delete[](void* ptr) {
    mgr_d.releaseBlock(ptr);
  }

  // method: setGrowSize
  //
  static bool8 setGrowSize(int32 grow_size) {
    return mgr_d.setGrow(grow_size);
  }

  // other memory management methods
  //
  bool8 clear(Integral::CMODE ctype = Integral::DEF_CMODE);
 
  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  set methods
  //
  //---------------------------------------------------------------------------
public:

  // method: setAlgorithm
  //
  bool8 setAlgorithm(ALGORITHM algorithm) {
    algorithm_d = algorithm;
    is_valid_d = false;
    return true;
  }
  
  // method: setImplementation
  //
  bool8 setImplementation(IMPLEMENTATION implementation) {
    implementation_d = implementation;
    is_valid_d = false;
    return true;  
  }

  // method: set
  //
  bool8 set(ALGORITHM algorithm = DEF_ALGORITHM,
	      IMPLEMENTATION implementation = DEF_IMPLEMENTATION,
	      int32 num_iterations = DEF_NUM_ITERATIONS,
	      float32 alpha_min = DEF_ALPHA_MIN,
	      float32 alpha_max = DEF_ALPHA_MAX,
	      int32 num_neighbors = DEF_NUM_NEIGHBORS,
	      int32 num_eigens = DEF_NUM_EIGENS) {
    
    algorithm_d = algorithm;
    implementation_d = implementation;
    num_iterations_d = num_iterations;
    alpha_min_d = alpha_min;
    alpha_max_d = alpha_max;
    num_neighbors_d = num_neighbors;
    num_eigens_d = num_eigens;

    return true;
  }

  // method: setNumIterations
  //
  bool8 setNumIterations(int32 num_iterations) {
    num_iterations_d.assign(num_iterations);
    return true;
  }

  // method: setMinAlpha
  //
  bool8 setMinAlpha(float32 alpha_min) {
    alpha_min_d.assign(alpha_min);
    return true;
  }

  // method: setMaxAlpha
  //
  bool8 setMaxAlpha(float32 alpha_max) {
    alpha_max_d.assign(alpha_max);
    return true;
  }

  // method: setNumNeighbors
  //
  bool8 setNumNeighbors(int32 num_neighbors) {
    num_neighbors_d.assign(num_neighbors);
    return true;
  }

  // method: setNumEigens
  //
  bool8 setNumEigens(int32 num_eigens) {
    num_eigens_d.assign(num_eigens);
    return true;
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods
  //  get methods
  //
  //---------------------------------------------------------------------------

  // method: getAlgorithm
  //
  ALGORITHM getAlgorithm() const {
    return algorithm_d;
  }
  
  // method: getImplementation
  //
  IMPLEMENTATION getImplementation() const {
    return implementation_d;
  }

  // method: get
  //
  bool8 get(ALGORITHM& algorithm,
	      IMPLEMENTATION& implementation,
	      int32& num_iterations,
	      float32& alpha_min,
	      float32& alpha_max,
	      int32& num_neighbors,
	      int32& num_eigens) {
    algorithm = algorithm_d;
    implementation = implementation_d;
    num_iterations = num_iterations_d;
    alpha_min = alpha_min_d;
    alpha_max = alpha_max_d;
    num_neighbors = num_neighbors_d;
    num_eigens = num_eigens_d;
    return true;
  }

  // method: getNumIterations
  //
  int32 getNumIterations() {
    return num_iterations_d;
  }

  // method: getMinAlpha
  //
  float32 getMinAlpha() {
    return alpha_min_d;
  }

  // method: getMaxAlpha
  //
  float32 getMaxAlpha() {
    return alpha_max_d;
  }

  // method: getNumNeighbors
  //
  int32 getNumNeighbors() {
    return num_neighbors_d;
  }

  // method: getNumEigens
  //
  int32 getNumEigens() {
    return num_eigens_d;
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  computational methods
  //
  //---------------------------------------------------------------------------

  // compute method when input is given as a matrix
  //
  bool8 compute(MatrixFloat& output, const MatrixFloat& input,
		  AlgorithmData::COEF_TYPE coef_type = DEF_COEF_TYPE,
		  int32 channel_index = DEF_CHANNEL_INDEX);
 
  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  public methods required by the AlgorithmBase interface contract
  //
  //---------------------------------------------------------------------------

  // assign method
  //
  bool8 assign(const AlgorithmBase& arg);

  // equality method
  //
  bool8 eq(const AlgorithmBase& arg) const;

  // method: className
  //
  const String& className() const {
    return CLASS_NAME;
  }

  // no initialization method is needed since none of the
  // algorithms currently supported require a history

  // initialization method
  //
  //bool8 init();
  
  // apply method
  //
  bool8 apply(Vector<AlgorithmData>& output,
		const Vector< CircularBuffer<AlgorithmData> >& input);

  // method to get the parser
  //
  bool8 setParser(SofParser* parser);

  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:

  // common i/o methods
  //
  bool8 readDataCommon(Sof& sof, const String& pname,
			 int32 size = SofParser::FULL_OBJECT,
			 bool8 param = true, bool8 nested = false);
  bool8 writeDataCommon(Sof& sof, const String& pname) const;

  //---------------------------------------------------------------------------
  //
  // protected methods
  //
  //---------------------------------------------------------------------------
protected:

  // algorithm-specific computation methods: 
  //

  // method: findNearestNeighbors
  //
  bool8 findNearestNeighbors(MatrixFloat& output, const MatrixFloat& input, 
			       int32 ref_state);

  // method: computeSvdClean
  //
  bool8 computeSvdClean(MatrixFloat& output, const MatrixFloat& input);
};

// end of include file
//
#endif
