// file: $isip_ifc/class/algo/DynamicFilter/lyap_05.cc
// version: $Id: lyap_05.cc 10636 2007-01-26 22:18:09Z tm334 $
//

// isip include files
//
#include "DynamicFilter.h"

// method: apply
//
// arguments:
//  Vector<AlgorithmData>& output: (output) output data
//  const Vector< CircularBuffer<AlgorithmData> >& input: (input) input data
//
// return: a bool8 value indicating status
//
// this method calls computation method with appropriate input and
// output data types
//
bool8 DynamicFilter::apply(Vector<AlgorithmData>& output_a,
			const Vector< CircularBuffer<AlgorithmData> >& input_a) {

  // determine the number of input channels and force the output to be
  // that number
  //
  int32 len = input_a.length();
  output_a.setLength(len);
  bool8 res = true;

  frame_index_d++;
  
  // start the debugging output
  //
  displayStart(this);

  // loop over the channels and call the compute method
  //
  for (int32 c = 0; c < len; c++) {

    // display the channel number
    //
    displayChannel(c);

    // call AlgorithmData::makeMatrixFloat to force the output vector for
    // this channel to be a MatrixFloat, call AlgorithmData::getVectorFloat
    // on the input for this channel to check that the input is
    // already a VectorFloat and return that vector.
    //
    if (input_a(c)(0).getDataType() == AlgorithmData::MATRIX_FLOAT) {
      
      res &= compute(output_a(c).makeMatrixFloat(),
		     input_a(c)(0).getMatrixFloat(),
		     AlgorithmData::RPS, c);
    }

    // error unknown coefficient type
    //
    else {
      return Error::handle(name(), L"apply", ERR_UNCTYP,
			   __FILE__, __LINE__);
    }
    
    // set the coefficient type for the output
    //    
    output_a(c).setCoefType(AlgorithmData::LYAPUNOV);
  }

  // possibly display the data
  //
  display(output_a, input_a, name());
    
  // finish the debugging output
  //
  displayFinish(this);

  // exit gracefully
  //
  return res;
}

// method: compute
//
// arguments:
//  MatrixFloat& output: (output) filtered RPS matrix
//  const MatrixFloat& input: (input) original RPS matrix
//  int32 index: (input) channel number
//
// return: a bool8 value indicating status
//
// this method filters the RPS using dynamic filtering
//
bool8 DynamicFilter::compute(MatrixFloat& output_a,
			       const MatrixFloat& input_a,
			       AlgorithmData::COEF_TYPE input_coef_type_a,
			       int32 index_a) {
  
  if (algorithm_d == SVD_CLEAN) {
    int32 num_rows = input_a.getNumRows();
    if (num_rows > (int32)0) {
      computeSvdClean(output_a, input_a);
    }
  }
  else {
    return Error::handle(name(), L"compute", ERR_UNKALG, __FILE__, __LINE__);
  }

  // exit gracefully
  //  
  return true;
}

// method: computeSvdClean
//
// arguments:
//  MatrixFloat& output: (output) filtered RPS matrix
//  const MatrixFloat& input: (input) input RPS matrix
//
// return: a bool8 value indicating status
//
// uses a Dynamic filtering method based on the SVD of the neighborhood matrix
// to clean the attractor
//
bool8 DynamicFilter::computeSvdClean(MatrixFloat& output_a, 
				       const MatrixFloat& input_a) {

  float32 alpha_incr = ((float32)alpha_max_d-(float32)alpha_min_d)/((float32)num_iterations_d-1);
  
  VectorFloat temp_column;
  MatrixFloat filtered_data, nearest_neighbors;

  output_a.copy(input_a);

  // remove the continuity column from the RPS before processing
  //
  output_a.getColumn(temp_column, 
		    output_a.getNumColumns()-1);
  
  output_a.setDimensions(output_a.getNumRows(),
			output_a.getNumColumns()-1,
			true);
  
  filtered_data.copy(output_a);
  
  float32 alpha = alpha_min_d;
  for (int32 iter=0; iter<num_iterations_d; iter++) {
    for (int32 i=0; i<output_a.getNumRows(); i++) {

      // starting the bulk of the algorithm
      //
      // get the nearest neighbors for the first point
      //
      findNearestNeighbors(nearest_neighbors, output_a, i);
      //nearest_neighbors.debug(L"nearest_neighbors");

      // do svd decomposition for the neighborhood matrix
      //
      MatrixFloat u, w, v;
      nearest_neighbors.decompositionSVD(u, w, v);
      
      //u.debug(L"u");
      //w.debug(L"w");
      //v.debug(L"v");

      // project the reference point onto the subspace spanned by the first
      // p eigen vectors of the nbhd matrix
      //
      VectorFloat projected_point;
      output_a.getRow(projected_point, i);
      
      VectorFloat eigen;
      v.getColumn(eigen,0);
      //eigen.debug(L"eigen");
      //projected_point.debug(L"projected_point");
      //projected_point.debug(L"projected_point");      
      eigen.mult(projected_point.dotProduct(eigen));
      projected_point = eigen;
      //projected_point.debug(L"projected_point");
      //return 0;            
      for (int j=1; j<num_eigens_d; j++) {
	
	VectorFloat temp_point;
	output_a.getRow(temp_point, i);
	v.getColumn(eigen, j);
	
	temp_point.mult(eigen);
	temp_point.mult(eigen);
	
	projected_point.add(temp_point);
      }
      
      //projected_point.debug(L"projected_point");
      
      VectorFloat shifted_point;
      output_a.getRow(shifted_point, i);    
      shifted_point.mult(1-alpha);
      projected_point.mult(alpha);
      shifted_point.add(projected_point);
      
      //shifted_point.debug(L"shifted_point");
      filtered_data.setRow(i,shifted_point);
    }
    
    output_a.copy(filtered_data);
    //cout << "alpha: " << alpha << endl;
    alpha += alpha_incr;
  }

  // add the continuity column back to the output matrix 
  //
  output_a.setDimensions(output_a.getNumRows(), output_a.getNumColumns()+1, true);
  output_a.setColumn(output_a.getNumColumns()-1, temp_column);

  // exit gracefully
  //
  return true;

}


// method: findNearestNeighbors
//
// arguments:
//  MatrixFloat& output: (output) neighborhood matrix
//  const MatrixFloat& input: (input) input RPS matrix
//  int32 ref_state: (input) reference state (center of neighborhood)
//
// return: a bool8 value indicating status
//
// finds the nearest neighbors to a given reference state
//
bool8 DynamicFilter::findNearestNeighbors(MatrixFloat& output_a, 
					    const MatrixFloat& input_a,
					    int32 ref_state_a) {

  output_a.setDimensions(num_neighbors_d, input_a.getNumColumns(), false);
  
  // given ref_state_a, need to compute distance between state and all other points
  // on the attractor. Then we need to determine which states are in the 
  // 'neighborhood' of the reference state
  //
  VectorFloat ref_point;
  input_a.getRow(ref_point, ref_state_a);
  Vector<Pair<Float, Long> > neighborhood(num_neighbors_d);
  
  // compute first num_neighbors_d distances and seed the neighborhood
  // matrix
  //
  for (int32 i=0; i<num_neighbors_d; i++) {
    Pair<Float, Long> tmp_neighbor;
    tmp_neighbor.second() = i;
    
    // if the reference state is in the first num_neighbors_d points, just
    // assign the distance some really big value
    //
    if (i == ref_state_a) {
      tmp_neighbor.first() = DYNAMIC_FILTER_INF;
    }
    else {
      VectorFloat tmp_point;
      input_a.getRow(tmp_point, i);
      //ref_point.debug(L"ref_point");
      //tmp_point.debug(L"tmp_point");
      
      tmp_neighbor.first() = ref_point.distance(tmp_point);
    }
    neighborhood(i) = tmp_neighbor;
  }

  // loop over all states and compute the distance
  //  
  for (int32 i=num_neighbors_d; i<input_a.getNumRows(); i++) {
    
    // don't consider the reference point as a part of the neighborhood
    //
    if (i != ref_state_a) {    
      
      Pair<Float, Long> tmp_neighbor;
      tmp_neighbor.second() = i;
      
      VectorFloat tmp_point;
      input_a.getRow(tmp_point, i);
      
      tmp_neighbor.first() = ref_point.distance(tmp_point);
      //tmp_point.sub(ref_point);
      //tmp_neighbor.first() = tmp_point.maxMag();
      
      // need to detemine whether or not this point could possibly be in
      // the neighborhood.
      //
      float32 max_distance = 0;
      int32 max_distance_index = 0;
      for (int j=0; j<num_neighbors_d; j++) {	
	if (neighborhood(j).first() > max_distance) {
	  max_distance = neighborhood(j).first();
	  max_distance_index = j;
	}
      }
      
      if (tmp_neighbor.first() < max_distance) {
	neighborhood(max_distance_index) = tmp_neighbor;
      }
    }
  }
  
  for (int32 i=0; i<num_neighbors_d; i++) {
    VectorFloat neighborhood_point;
    input_a.getRow(neighborhood_point, neighborhood(i).second());
    output_a.setRow(i,neighborhood_point);
  }
  
  // exit gracefully
  //  
  return true;
}
