// file: $isip_ifc/class/algo/KalmanFilter/kf_03.cc
// version: $Id: kf_03.cc 10602 2006-08-17 17:43:12Z ss754 $
//

// isip include files
//
#include "KalmanFilter.h"

//------------------------------------------------------------------
// these methods have to be in the same file so they can use the same
// static parser pointer
//------------------------------------------------------------------

// declare a static sof parser pointer
//
static SofParser* parser_d = (SofParser*)NULL;

// method: setParser
//
// arguments:
//  SofParser* parser: (input) sof file object
//
// return: a bool8 value indicating status
//
// this method sets the parser from the parent object to be used in
// the next readData method.
//
bool8 KalmanFilter::setParser(SofParser* parser_a) {

  // set the parser
  //
  parser_d = parser_a;

  // exit gracefully
  //
  return true;
}

// method: read
//
// arguments:
//  Sof& sof: (input) sof file object
//  int32 tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a bool8 value indicating status
//
// this method has the object read itself from an Sof file according
// to the specified name and tag
//
bool8 KalmanFilter::read(Sof& sof_a, int32 tag_a, const String& name_a) {

  // read the instance of the object from the Sof file
  //
  if (!sof_a.find(name_a, tag_a)) {
    return false;
  }

  // read the actual data from the sof file
  //
  return readData(sof_a);
}

// method: readData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  int32 size: (input) size in bytes of object (or FULL_OBJECT)
//  bool8 param: (input) is the parameter name in the file?
//  bool8 nested: (input) are we nested?
//
// return: a bool8 value indicating status
//
// this method has the object read itself from an Sof file. it assumes
// that the Sof file is already positioned correctly.
//
bool8 KalmanFilter::readData(Sof& sof_a, const String& pname_a, int32 size_a,
			   bool8 param_a, bool8 nested_a) {
  
  // declare local variables
  //
  bool8 status = true;

  // we need a parser
  // 
  if (parser_d == (SofParser*)NULL) {
    
    parser_d = new SofParser;
    
    if (nested_a) {
      parser_d->setNest();
    }
    
    // load the parse
    //
    if (!parser_d->load(sof_a, size_a)) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
                           __FILE__, __LINE__, Error::WARNING);
    }           
  }

  // set the debug level of the parser
  //
  parser_d->setDebug(debug_level_d.getIndex());
  
  // get the algorithm name
  //
  if (parser_d->isPresent(sof_a, PARAM_ALGORITHM)) {
    if (!ALGO_MAP.readElementData((int32&)algorithm_d, sof_a,
				  PARAM_ALGORITHM,
				  parser_d->getEntry(sof_a, PARAM_ALGORITHM))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    algorithm_d = DEF_ALGORITHM;
  }

  // get the implementation
  //
  if (parser_d->isPresent(sof_a, PARAM_IMPLEMENTATION)) {
    if (!IMPL_MAP.readElementData((int32&)implementation_d, sof_a,
				  PARAM_IMPLEMENTATION,
				  parser_d->getEntry(sof_a,
						     PARAM_IMPLEMENTATION))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    implementation_d = DEF_IMPLEMENTATION;
  }
  
  // get the debug level
  //
  if (parser_d->isPresent(sof_a, PARAM_DBGL)) {
    if (!debug_level_d.readData(sof_a, PARAM_DBGL,
				parser_d->getEntry(sof_a, PARAM_DBGL))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    debug_level_d = Integral::DEF_DEBUG;
  }

  // check known algorithms: CONVENTIONAL
  //
  if ((algorithm_d == CONVENTIONAL) && (implementation_d == AR_SCALAR)) {
    status = readDataKalmanFilter(sof_a, pname_a, size_a, param_a, nested_a);
  }

  // check unsupported algorithms
  //
  else {
    return Error::handle(name(), L"readData",
			 ERR_UNKALG, __FILE__, __LINE__);
  }

  // check that all parameters are accounted for
  //
  if (!parser_d->checkParams(sof_a)) {

    // delete the parser
    //
    delete parser_d;
    parser_d = (SofParser*)NULL;
    
    // return a warning message
    //
    return Error::handle(name(), L"readData", Error::IO,
			 __FILE__, __LINE__, Error::WARNING);
  }

  // delete the parser
  //
  delete parser_d;
  parser_d = (SofParser*)NULL;

  // exit gracefully
  //
  return status;
}

// method: readDataKalmanFilter
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  int32 size: (input) size in bytes of object (or FULL_OBJECT)
//  bool8 param: (input) is the parameter name in the file?
//  bool8 nested: (input) are we nested?
//
// return: a bool8 value indicating status
//
// this method assumes the algorithm has been read, and proceeds to read
// the rest of the object for the AR_SCALAR algorithm
//
bool8 KalmanFilter::readDataKalmanFilter(Sof& sof_a, const String& pname_a,
			     int32 size_a, bool8 param_a,
			     bool8 nested_a) {
  // get the order
  //
  if (parser_d->isPresent(sof_a, PARAM_ORDER)) {
    if (!order_d.readData(sof_a, PARAM_ORDER,
			  parser_d->getEntry(sof_a, PARAM_ORDER))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    order_d = DEF_ORDER;
  }

  // get the state mem
  //
  if (parser_d->isPresent(sof_a, PARAM_STATE_MEM)) {
    if (!state_mem_d.readData(sof_a, PARAM_STATE_MEM,
			  parser_d->getEntry(sof_a, PARAM_STATE_MEM))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    state_mem_d = DEF_STATE_MEM;
  }

  // get trn
  //
  if (parser_d->isPresent(sof_a, PARAM_TRN)) {
    if (!trn_d.readData(sof_a, PARAM_TRN,
			  parser_d->getEntry(sof_a, PARAM_TRN), false, true)) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    trn_d = DEF_TRN;
  }

  // get obs
  //
  if (parser_d->isPresent(sof_a, PARAM_OBS)) {
    if (!obs_d.readData(sof_a, PARAM_OBS,
			  parser_d->getEntry(sof_a, PARAM_OBS),
			false, false)) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    obs_d = DEF_OBS;
  }

  // get process noise covar
  //
  if (parser_d->isPresent(sof_a, PARAM_NOISE_PRC)) {
    if (!covar_nse_pro_d.readData(sof_a, PARAM_NOISE_PRC,
			  parser_d->getEntry(sof_a, PARAM_NOISE_PRC),
				  false, true)) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    covar_nse_pro_d = DEF_NOISE_PRC;
  }

  // get obs noise
  //
  if (parser_d->isPresent(sof_a, PARAM_NOISE_OBS)) {
    if (!covar_nse_obs_d.readData(sof_a, PARAM_NOISE_OBS,
			  parser_d->getEntry(sof_a, PARAM_NOISE_OBS),
				  false, false)) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    covar_nse_obs_d = DEF_NOISE_OBS;
  }

  // get estimated error covar
  //
  if (parser_d->isPresent(sof_a, PARAM_NOISE_ERR)) {
    if (!covar_err_est_d.readData(sof_a, PARAM_NOISE_ERR,
			  parser_d->getEntry(sof_a, PARAM_NOISE_ERR),
				  false, true)) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataKalmanFilter", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    covar_err_est_d = DEF_NOISE_ERR;
  }
  
  // exit gracefully
  //
  return true;
}
