// file: $isip_ifc/class/algo/Lyapunov/lyap_04.cc
// version: $Id: lyap_04.cc 10485 2006-03-15 17:22:34Z srinivas $
//

// isip include files
//
#include "Lyapunov.h"

// method: sofSize
//
// arguments: none
//
// return: size of object
//
// this method returns the size of the object in the Sof file and is
// used for binary write
//
int32 Lyapunov::sofSize() const {
  
  // start with the space required for the algorithm name
  //
  int32 bytes = ALGO_MAP.elementSofSize();

  // add the space required for the implementation name
  //
  bytes += IMPL_MAP.elementSofSize();

  // add the space required for the number of neighbors
  //
  bytes += num_neighbors_d.sofSize();

  // add the space required for the evolve step
  //  
  bytes += evolve_step_d.sofSize();

  // add the space required for the local dimension
  //  
  bytes += local_dim_d.sofSize();
  
  // add the space required for the number of steps
  //
  bytes += num_steps_d.sofSize();

  // add the space required for inner radius
  //  
  bytes += inner_radius_d.sofSize();

  // add the space required for outer radius
  //  
  bytes += outer_radius_d.sofSize();

  // add the space required for number of neighbor subgroups
  //  
  bytes += num_neighbor_subgroups_d.sofSize();
  
  // add the space required for reinitialization step
  //
  bytes += reinit_step_d.sofSize();

  // add the space required for the start position
  //
  bytes += start_pos_d.sofSize();

  // add the space required for the debug level
  //
  bytes += debug_level_d.sofSize();
  
  // return the size
  //
  return bytes;
}

// method: write
//
// arguments:
//  Sof& sof: (input) sof file object
//  int32 tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a bool8 value indicating status
//
// this method has the object write itself to an Sof file
//
bool8 Lyapunov::write(Sof& sof_a, int32 tag_a,
			  const String& name_a) const {

  // declare local variable
  //
  int32 obj_size;

  if (sof_a.isText()) {

    // set the size to be dynamic
    //
    obj_size = Sof::ANY_SIZE;
  }
  else {
    
    // set the size to be the size of the object written to the Sof file
    //
    obj_size = sofSize();
  }
  
  // write the object into the sof file's index
  //
  if (!sof_a.put(name_a, tag_a, obj_size)) {
    return false;
  }

  // write data and exit gracefully
  //
  return writeData(sof_a);
}

// method: writeData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//
// return: a bool8 value indicating status
//
// this method branches on the algorithm name
//
bool8 Lyapunov::writeData(Sof& sof_a, const String& pname_a) const {

  // declare local variables
  //
  bool8 status = false;
  
  // write a start string if necessary
  //
  sof_a.writeLabelPrefix(pname_a);

  // write the algorithm name
  //
  ALGO_MAP.writeElementData(sof_a, PARAM_ALGORITHM, (int32)algorithm_d);

  // invoke the proper lower-level method
  //
  if (algorithm_d == LINEAR_TANGENT_MAP) {
    status = writeDataCommon(sof_a, pname_a);
  }

  // exit ungracefully
  //
  else {
    return Error::handle(name(), L"writeData",
			 ERR_UNKALG, __FILE__, __LINE__);
  }

  // put an end string if necessary
  //
  sof_a.writeLabelSuffix(pname_a);
  
  // exit gracefully
  //
  return status;
}

// method: writeDataCommon
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//
// return: a bool8 value indicating status
//
// this method writes data to a file for several algorithms that
// share the same parameters
//
bool8 Lyapunov::writeDataCommon(Sof& sof_a,
				    const String& pname_a) const {

  // write the implementation type
  //
  IMPL_MAP.writeElementData(sof_a, PARAM_IMPLEMENTATION,
			    (int32)implementation_d);

  // write the num of neighbors
  //
  num_neighbors_d.writeData(sof_a, PARAM_NUM_NEIGHBORS);

  // write the evolve steps
  //
  evolve_step_d.writeData(sof_a, PARAM_EVOLVE_STEP);

  // write the local dimension
  //
  local_dim_d.writeData(sof_a, PARAM_LOCAL_DIM);
  
  // write the number of steps
  //
  num_steps_d.writeData(sof_a, PARAM_NUM_STEPS);

  // write the inner radius
  //
  inner_radius_d.writeData(sof_a, PARAM_INNER_RADIUS);

  // write the outer radius
  //
  outer_radius_d.writeData(sof_a, PARAM_OUTER_RADIUS);

  // write the number of subgroups of neighbors
  //
  num_neighbor_subgroups_d.writeData(sof_a, PARAM_NUM_NEIGHBOR_SUBGROUPS);

  // write the reinitialization step
  //
  reinit_step_d.writeData(sof_a, PARAM_REINIT_STEP);

  // write the starting point for LE calculation
  //
  start_pos_d.writeData(sof_a, PARAM_START_POS);

  // write the debug level (from the base class)
  //
  debug_level_d.writeData(sof_a, PARAM_DBGL);

  // exit gracefully
  //
  return true;
}
