// file: $isip/class/algo/Math/math_09.cc
// version: $Id: math_09.cc 8037 2002-04-01 20:04:33Z gao $
//

// isip include files
//
#include "Math.h"

// method: computeFuncCalcEnumerateCFloat
//
// arguments:
//  AlgorithmData& output: (output) output data
//  const Vector<AlgorithmData>& input: (input) input data
//
// return: a bool8 value indicating status
//
// this method implements following function:
//   Y = operation(0) {a(0)*function(0)(X(0))} operation(1)
//   {a(1)*function(1)(X(1))} operation(2) {a(2)*function(2)(X(2))}...
//   ... + const,
// where X(0), X(1), ... are the parts of input VectorComplexFloat splitted
// by number of operands. These can be VectorComplexFloat. If
// user wants to implement this for scalar or single vector, then
// number of operands will be 1. For scalar input the length of the
// input vector will be 1.
//
bool8 Math::computeFuncCalcEnumerateCFloat(AlgorithmData& output_a,
					     const Vector<AlgorithmData>&
					     input_a) {

  // check algorithm and implementation names
  //
  if ((algorithm_d != FUNCTION_CALCULATOR) ||
      (implementation_d != ENUMERATE)) {
    return Error::handle(name(), L"computeFuncCalcEnumerateCFloat", Error::ARG,
                         __FILE__, __LINE__);
  }

  // compute the number of combinations
  //
  if (num_operands_d != input_a.length()) {
    return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			 Error::ARG, __FILE__, __LINE__);
  }

  // input should be either vector or matrix
  //
  for (int32 j = 0; j < num_operands_d; j++) {
    if ((input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) &&
	(input_a(j).getDataType() != AlgorithmData::MATRIX_COMPLEX_FLOAT)) {
      return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			   Error::ARG, __FILE__, __LINE__);
    }
  }

  // special case for one complex operation such as mag, real and imag
  //
  if (num_operands_d == (int32)1) {
    if (function_d(0) == (int32)MAG) {
      input_a(0).getVectorComplexFloat().mag(output_a.makeVectorFloat());
      output_a.getVectorFloat().add((float32)const_d);
      return true;
    }
    else if (function_d(0) == (int32)REAL) {
      input_a(0).getVectorComplexFloat().real(output_a.makeVectorFloat());
      output_a.getVectorFloat().add((float32)const_d);
      return true;
    }
    else if (function_d(0) == (int32)IMAG) {
      input_a(0).getVectorComplexFloat().imag(output_a.makeVectorFloat());
      output_a.getVectorFloat().add((float32)const_d);
      return true;
    }
  }

   
  // set the mode of output
  //
  if (input_a(0).getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
    output_a.makeVectorComplexFloat();
  }
  else {
    output_a.makeMatrixComplexFloat();
  }

  // loop over the number of input operands
  //
  for (int32 j = 0; j < num_operands_d; j++) {
    
    // declare temporary input variable
    //
    AlgorithmData tmp_input;
    
    // set the type to either matrix or vector, depending upon the
    // type of the input
    //
    if (input_a(j).getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
      tmp_input.makeVectorComplexFloat().assign(input_a(j).getVectorComplexFloat());
    }
    else {
      tmp_input.makeMatrixComplexFloat().assign(input_a(j).getMatrixComplexFloat());
    }

    // branch on function:
    //  apply function and corresponding weight
    //
    if (function_d(j) == (int32)IDENTITY) {
      
      // do nothing
      //
    }
    
    // function: EXP
    //
    else if (function_d(j) == (int32)EXP) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().exp();
    }
    
    // function: EXP2
    //    
    else if (function_d(j) == (int32)EXP2) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().exp2();
    }
    
    // function: EXP10
    //    
    else if (function_d(j) == (int32)EXP10) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().exp10();
    }
    
    // function: FACTORIAL
    //    
    else if (function_d(j) == (int32)FACTORIAL) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().factorial();
    }
    
    // function: LOG
    //        
    else if (function_d(j) == (int32)LOG) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
  
      // check vector elements
      //
      for (int32 index =  tmp_input.getVectorComplexFloat().length() - 1;
	   index >= 0; index--) {
	if (tmp_input.getVectorComplexFloat()(index) <= (float32)0.0 ) {
	  return Error::handle(name(), L"computeFuncCalcEnumerateCFloat", Error::ARG, __FILE__, __LINE__);
	}
      }
      tmp_input.getVectorComplexFloat().log();
    }
    
    // function: LOG2
    //            
    else if (function_d(j) == (int32)LOG2) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      
      // check vector elements
      //
      for (int32 index =  tmp_input.getVectorComplexFloat().length() - 1;
	   index >= 0; index--) {
	if (tmp_input.getVectorComplexFloat()(index) <= (float32)0.0 ) {
	  return Error::handle(name(), L"computeFuncCalcEnumerateCFloat", Error::ARG, __FILE__, __LINE__);
	}
      }
      tmp_input.getVectorComplexFloat().log2();
    }
    
    // function: LOG10
    //            
    else if (function_d(j) == (int32)LOG10) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      
      // check vector elements
      //
      for (int32 index =  tmp_input.getVectorComplexFloat().length() - 1;
	   index >= 0; index--) {
	if (tmp_input.getVectorComplexFloat()(index) <= (float32)0.0 ) {
	  return Error::handle(name(), L"computeFuncCalcEnumerateCFloat", Error::ARG, __FILE__, __LINE__);
	}
      }
      tmp_input.getVectorComplexFloat().log10();
    }
    
    // function: LOG1P
    //            
    else if (function_d(j) == (int32)LOG1P) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      
      // check vector elements
      //
      for (int32 index =  tmp_input.getVectorComplexFloat().length() - 1;
	   index >= 0; index--) {
	if (tmp_input.getVectorComplexFloat()(index) <= (float32)0.0 ) {
	  return Error::handle(name(), L"computeFuncCalcEnumerateCFloat", Error::ARG, __FILE__, __LINE__);
	}
      }
      tmp_input.getVectorComplexFloat().log1p();
    }
    
    // function: ABS
    //            
    else if (function_d(j) == (int32)ABS) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			   ERR_UNKFUN, __FILE__, __LINE__);
    }
    
    // function: NEG
    //            
    else if (function_d(j) == (int32)NEG) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	tmp_input.getVectorComplexFloat().neg();
      }
      else if (input_a(j).getDataType() != AlgorithmData::MATRIX_COMPLEX_FLOAT) {
	tmp_input.getMatrixComplexFloat().neg();
      }
      else {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     Error::ARG, __FILE__, __LINE__);
      }
    }
    
    // function: ROUND
    //                 
    else if (function_d(j) == (int32)ROUND) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().round();
    }
    
    // function: CEIL
    //                
    else if (function_d(j) == (int32)CEIL) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().ceil();
    }
    
    // function: FLOOR
    //                    
    else if (function_d(j) == (int32)FLOOR) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().floor();
    }
    
    // function: RFLOOR
    //                        
    else if (function_d(j) == (int32)RFLOOR) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().rfloor();
    }
    
    // function: SIN
    //                        
    else if (function_d(j) == (int32)SIN) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().sin();
    }
    
    // function: COS
    //                            
    else if (function_d(j) == (int32)COS) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().cos();
    }

    // function: TAN
    //                            
    else if (function_d(j) == (int32)TAN) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().tan();
    }
    
    // function: ASIN
    //                            
    else if (function_d(j) == (int32)ASIN) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().asin();
    }
    
    // function: ACOS
    //                            
    else if (function_d(j) == (int32)ACOS) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().acos();
    }
    
    // function: ATAN
    //                            
    else if (function_d(j) == (int32)ATAN) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().atan();
    }
    
    // function: SQUARE
    //                            
    else if (function_d(j) == (int32)SQUARE) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().square();
    }
    
    // function: SQRT
    //                            
    else if (function_d(j) == (int32)SQRT) {
      if (input_a(j).getDataType() != AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getVectorComplexFloat().sqrt();
    }
    
    // function: INVERSE
    //                            
    else if (function_d(j) == (int32)INVERSE) {
      if (input_a(j).getDataType() != AlgorithmData::MATRIX_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getMatrixComplexFloat().inverse();
    }
    
    // function: TRANSPOSE
    //                            
    else if (function_d(j) == (int32)TRANSPOSE) {
      if (input_a(j).getDataType() != AlgorithmData::MATRIX_COMPLEX_FLOAT) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_FNDTYP, __FILE__, __LINE__);
      }
      tmp_input.getMatrixComplexFloat().transpose();
    }
    
    // invalid function
    //
    else {
      return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			   ERR_UNKFUN, __FILE__, __LINE__);
    }
    
    // apply weight
    //
    if (input_a(j).getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
      tmp_input.getVectorComplexFloat().mult(complexfloat(weight_d(j)));
    }
    else if (input_a(j).getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {
      tmp_input.getMatrixComplexFloat().mult(complexfloat(weight_d(j)));
    }
    else {
      return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			   Error::ARG, __FILE__, __LINE__);
    }

    // branch on operation:
    //  Operation: ASSIGN
    //
    if (operation_d(j) == (int32)ASSIGN) {
      
      // this operation is possible between same data types
      //
      if (output_a.getDataType() != input_a(j).getDataType()) {
	return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			     ERR_MATCH, __FILE__, __LINE__);
      }
      if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
	output_a.makeVectorComplexFloat().assign(tmp_input.getVectorComplexFloat());
      }
      else {
	output_a.makeMatrixComplexFloat().assign(tmp_input.getMatrixComplexFloat());
      }
    }
    
    //  Operation: ADD
    //
    else if (operation_d(j) == (int32)ADD) {
      
      computeAddCFloat(output_a, tmp_input);
    }

    //  Operation: SUBTRACT
    //
    else if (operation_d(j) == (int32)SUBTRACT) {
      
      computeSubCFloat(output_a, tmp_input);
    }

    //  Operation: MULTIPLY
    //
    else if (operation_d(j) == (int32)MULTIPLY) {

      computeMultCFloat(output_a, tmp_input);
    }
    
    //  Operation: DIVIDE
    //
    else if (operation_d(j) == (int32)DIVIDE) {
      
      computeDivCFloat(output_a, tmp_input);
    }
    
    // invalid operation
    //
    else {
      return Error::handle(name(), L"computeFuncCalcEnumerateCFloat",
			   ERR_UNKOPE, __FILE__, __LINE__);
    }
  }

  // add the constant term
  //
  if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
    output_a.getVectorComplexFloat().add((complexfloat)const_d);
  }
  else {
    output_a.getMatrixComplexFloat().add((complexfloat)const_d);
  }
  
  // exit gracefully
  //
  return true;
}

// method: computeAddCFloat
//
// arguments:
//  AlgorithmData& output: (input/output) 1st operand and output
//  const AlgorithmData& input: (input) 2nd operand
//
// return: logical error status
//
bool8 Math::computeAddCFloat(AlgorithmData& output_a,
			       const AlgorithmData& input_a) const {
  
  if (input_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
    
    int32 ilen = input_a.getVectorComplexFloat().length();

    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {

      int32 olen = output_a.getVectorComplexFloat().length();
      
      // if either input is of length 1, perform scalar addition
      //
      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getVectorComplexFloat().add(scalar);
      }
      else if (olen == 1) {
	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	return output_a.getVectorComplexFloat().add(input_a.getVectorComplexFloat(), scalar);
      }
      
      // if they are the same length, do vector addition
      //
      else if (ilen == olen) {
	return output_a.getVectorComplexFloat().add(input_a.getVectorComplexFloat());
      }

      // else error
      //
      else {
	return Error::handle(name(), L"computeAddCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }

    // 1 vector, 1 matrix. if the vector is of length 1, treat it as a
    // scalar
    //
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getMatrixComplexFloat().add(scalar);
      }
      else {
	return Error::handle(name(), L"computeAddCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }
  }

  else if (input_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

    // 1 vector, 1 matrix. if the vector is of length 1, treat it as a
    // scalar
    //
    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
      int32 olen = output_a.getVectorComplexFloat().length();
      if (olen == 1) {
	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	return output_a.makeMatrixComplexFloat().add(input_a.getMatrixComplexFloat(), scalar);
      }
      else {
	return Error::handle(name(), L"computeAddCFloat", ERR_MATCH, __FILE__,
			     __LINE__);
      }
    }

    // if both are matrices, just add them
    //
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {
      return output_a.getMatrixComplexFloat().add(input_a.getMatrixComplexFloat());
    }
  }

  // exit gracefully
  //
  return true;
}


// method: computeSubCFloat
//
// arguments:
//  AlgorithmData& output: (input/output) 1st operand and output
//  const AlgorithmData& input: (input) 2nd operand
//
// return: logical error status
//
bool8 Math::computeSubCFloat(AlgorithmData& output_a,
			       const AlgorithmData& input_a) const {
  
  if (input_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
    
    int32 ilen = input_a.getVectorComplexFloat().length();

    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {

      int32 olen = output_a.getVectorComplexFloat().length();
      
      // if either input is of length 1, perform scalar subtraction
      //
      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getVectorComplexFloat().sub(scalar);
      }
      else if (olen == 1) {

	// note this is different from addition since the operation is
	// not cummative
	//
	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	output_a.getVectorComplexFloat().setLength(ilen);
	output_a.getVectorComplexFloat().assign(scalar);
	return output_a.getVectorComplexFloat().sub(input_a.getVectorComplexFloat());
      }
      
      // if they are the same length, do vector subtraction
      //
      else if (ilen == olen) {
	return output_a.getVectorComplexFloat().sub(input_a.getVectorComplexFloat());
      }

      // else error
      //
      else {
	return Error::handle(name(), L"computeSubCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }

    // 1 vector, 1 matrix. if the vector is of length 1, treat it as a
    // scalar
    //
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getMatrixComplexFloat().sub(scalar);
      }
      else {
	return Error::handle(name(), L"computeSubCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }
  }

  else if (input_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

    // 1 vector, 1 matrix. if the vector is of length 1, treat it as a
    // scalar
    //
    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
      int32 olen = output_a.getVectorComplexFloat().length();
      if (olen == 1) {

	// note this is different from addition since the operation is
	// not cummative
	//
	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	output_a.makeMatrixComplexFloat().setDimensions(input_a.getMatrixComplexFloat());
	output_a.getMatrixComplexFloat().assign(scalar);
	return output_a.getMatrixComplexFloat().sub(input_a.getMatrixComplexFloat());
      }
      else {
	return Error::handle(name(), L"computeSubCFloat", ERR_MATCH, __FILE__,
			     __LINE__);
      }
    }

    // if both are matrices, just subtract them
    //
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {
      return output_a.getMatrixComplexFloat().sub(input_a.getMatrixComplexFloat());
    }
  }

  // exit gracefully
  //
  return true;
}

// method: computeMultCFloat
//
// arguments:
//  AlgorithmData& output: (input/output) 1st operand and output
//  const AlgorithmData& input: (input) 2nd operand
//
// return: logical error status
//
bool8 Math::computeMultCFloat(AlgorithmData& output_a,
				const AlgorithmData& input_a) const {
  
  if (input_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {
    
    int32 ilen = input_a.getVectorComplexFloat().length();

    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {

      int32 olen = output_a.getVectorComplexFloat().length();

      // if either input is of length 1, perform scalar multiplication
      //
      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getVectorComplexFloat().mult(scalar);
      }
      else if (olen == 1) {

	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	return output_a.getVectorComplexFloat().mult(input_a.getVectorComplexFloat(), scalar);
      }
      
      // if they are the same length, do vector multiplication
      //
      else if (ilen == olen) {
	return output_a.getVectorComplexFloat().mult(input_a.getVectorComplexFloat());
      }

      // else error
      //
      else {
	return Error::handle(name(), L"computeMultCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

      // if the vector is length 1, do scalar multiplication
      //
      if (ilen == 1) {
	complexfloat scalar = input_a.getVectorComplexFloat()(0);
	return output_a.getMatrixComplexFloat().mult(scalar);
      }

      // if the vector is the correct length for multv, do it
      //
      else if (ilen == output_a.getMatrixComplexFloat().getNumColumns()) {
	VectorComplexFloat tmp_out;
	output_a.getMatrixComplexFloat().multv(tmp_out, input_a.getVectorComplexFloat());
	output_a.makeVectorComplexFloat().assign(tmp_out);
      }

      // else error
      //
      else {
	return Error::handle(name(), L"computeMultCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }
  }

  // input is a matrix
  //
  else if (input_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

    if (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) {

      int32 olen = output_a.getVectorComplexFloat().length();

      // if the vector is of length 1, do scalar multiplication
      //
      if (olen == 1) {
	complexfloat scalar = output_a.getVectorComplexFloat()(0);
	output_a.makeMatrixComplexFloat().assign(input_a.getMatrixComplexFloat());
	return output_a.getMatrixComplexFloat().mult(scalar);
      }

      // if the vector is the correct length for vmult, do it
      //
      else if (olen == input_a.getMatrixComplexFloat().getNumRows()) {

	VectorComplexFloat tmp_out;
	input_a.getMatrixComplexFloat().vmult(tmp_out, output_a.getVectorComplexFloat());
	output_a.getVectorComplexFloat().assign(tmp_out);
      }
	
      // else error
      //
      else {
	return Error::handle(name(), L"computeMultCFloat", ERR_MATCH,
			     __FILE__, __LINE__);
      }
    }
    else if (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT) {

      return output_a.getMatrixComplexFloat().mult(input_a.getMatrixComplexFloat());
    }
  }

  // exit gracefully
  //
  return true;
}

// method: computeDivCFloat
//
// arguments:
//  AlgorithmData& output: (input/output) 1st operand and output
//  const AlgorithmData& input: (input) 2nd operand
//
// return: logical error status
//
bool8 Math::computeDivCFloat(AlgorithmData& output_a,
			       const AlgorithmData& input_a) const {  
  
  // vector-matrix division is possible only if length of the
  // vector is 1. no other matrix division is possible
  //
  if ((input_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) &&
      (output_a.getDataType() == AlgorithmData::MATRIX_COMPLEX_FLOAT)) {      

    if (input_a.getVectorComplexFloat().length() > 1) {
      return Error::handle(name(), L"computeDivCFloat",
			   ERR_MATCH, __FILE__, __LINE__);
    }
    VectorComplexFloat vec(input_a.getVectorComplexFloat());
    complexfloat scalar = vec(0);
    output_a.getMatrixComplexFloat().div(scalar);
  }

  // both inputs are vectors
  //
  else if ((input_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT) &&
	   (output_a.getDataType() == AlgorithmData::VECTOR_COMPLEX_FLOAT)) {

    // if the input is of length 1, do scalar division
    //
    if (input_a.getVectorComplexFloat().length() == 1) {
      complexfloat scalar = input_a.getVectorComplexFloat()(0);

      return output_a.getVectorComplexFloat().div(scalar);
    }

    // if the lengths are equal, do element-wise division
    //
    else if (input_a.getVectorComplexFloat().length() == output_a.getVectorComplexFloat().length()) {

      return output_a.getVectorComplexFloat().div(input_a.getVectorComplexFloat());

    } 
    else {
      return Error::handle(name(), L"computeDivCFloat", ERR_MATCH,
			   __FILE__, __LINE__);
    }
  }
  else {
    return Error::handle(name(), L"computeDivCFloat", ERR_MATCH, __FILE__, __LINE__);
  }
  
  // exit gracefully
  //
  return true;
}
