// file: $isip/class/algo/Prediction/Prediction.h
// version: $Id: Prediction.h 10636 2007-01-26 22:18:09Z tm334 $
//

// make sure definitions are only made once
//
#ifndef ISIP_PREDICTION
#define ISIP_PREDICTION

#ifndef ISIP_ALGORITHM_BASE
#include <AlgorithmBase.h>
#endif

#ifndef ISIP_CORRELATION
#include <Correlation.h>
#endif

#ifndef ISIP_COVARIANCE
#include <Covariance.h>
#endif

#ifndef ISIP_REFLECTION
#include <Reflection.h>
#endif

// Prediction: a class that computes the linear prediction
// coefficients from autocorrelation, covariance matrix, signal,
// reflection coefficients or log-area-ratios. four algorithms have
// been implemented: AUTOCORRELATION, COVARIANCE, LATTICE and
// REFLECTION.
//
class Prediction : public AlgorithmBase {

  //---------------------------------------------------------------------------
  //
  // public constants
  //
  //---------------------------------------------------------------------------
public:
  
  // define the class name
  //
  static const String CLASS_NAME;
  
  //----------------------------------------
  //
  // other important constants
  //
  //----------------------------------------
    
  // define the algorithm choices
  //
  enum ALGORITHM { AUTOCORRELATION = 0, COVARIANCE, LATTICE, REFLECTION,
		   LOG_AREA_RATIO, DEF_ALGORITHM = AUTOCORRELATION };
  
  // define the implementation choices
  //
  enum IMPLEMENTATION { DURBIN = 0, LEROUX_GUEGUEN, CHOLESKY, BURG,
			STEP_DOWN, KELLY_LOCHBAUM,
			DEF_IMPLEMENTATION = DURBIN };

  // define the static NameMap objects
  //
  static const NameMap ALGO_MAP;
  static const NameMap IMPL_MAP;

  //----------------------------------------
  //
  // i/o related constants
  //
  //----------------------------------------  
  
  static const String DEF_PARAM;
  static const String PARAM_ALGORITHM;
  static const String PARAM_IMPLEMENTATION;
  static const String PARAM_ORDER;
  static const String PARAM_DYN_RANGE;

  //----------------------------------------
  //
  // default values and arguments
  //
  //----------------------------------------  
  
  // define the default value(s) of the class data
  //
  static const int32 DEF_ORDER = -1;
  static const float DEF_DYN_RANGE = -100.0;

  // define default argument(s)
  //
  static const AlgorithmData::COEF_TYPE DEF_COEF_TYPE =
  AlgorithmData::CORRELATION;
  
  //----------------------------------------
  //
  // error codes
  //
  //----------------------------------------  
  
  static const int32 ERR = 71100;
  static const int32 ERR_DYNRANGE = 71101;
  static const int32 ERR_BETA = 71102;
  static const int32 ERR_ENERGY = 71103;
  static const int32 ERR_PREDERR = 71104;
  
  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:
  
  // algorithm name
  //
  ALGORITHM algorithm_d;

  // implementation name
  //
  IMPLEMENTATION implementation_d;

  // parameters related to the algorithm specification
  //
  Long order_d;
  Float dyn_range_d;
  
  // a static memory manager
  //
  static MemoryManager mgr_d;
  
  //---------------------------------------------------------------------------
  //
  // required public methods
  //
  //---------------------------------------------------------------------------
public:
    
  // method: name
  //
  static const String& name() {
    return CLASS_NAME;
  }

  // other static methods
  //
  static bool8 diagnose(Integral::DEBUG debug_level);
  
  // debug methods:
  //  setDebug is inherited from the base class
  //
  bool8 debug(const unichar* msg) const;

  // method: destructor
  //
  ~Prediction() {}

  // constructor
  //
  Prediction(ALGORITHM algorithm = DEF_ALGORITHM,
	     IMPLEMENTATION implementation = DEF_IMPLEMENTATION,
	     int32 order = DEF_ORDER,
	     float dyn_range = DEF_DYN_RANGE) {
    algorithm_d = algorithm;
    implementation_d = implementation;
    order_d = order;
    dyn_range_d = dyn_range;
    is_valid_d = false;
  }

  // method: copy constructor
  //
  Prediction(const Prediction& arg) {
    assign(arg);
  }
  
  // assign methods
  //
  bool8 assign(const Prediction& arg);

  // method: operator=
  //
  Prediction& operator= (const Prediction& arg) {
    assign(arg);
    return *this;
  }
  
  // i/o methods
  //
  int32 sofSize() const;
  
  bool8 read(Sof& sof, int32 tag, const String& name = CLASS_NAME);
  bool8 write(Sof& sof, int32 tag, const String& name = CLASS_NAME) const;

  bool8 readData(Sof& sof, const String& pname = DEF_PARAM,
                   int32 size = SofParser::FULL_OBJECT,
                   bool8 param = true,
                   bool8 nested = false);
  bool8 writeData(Sof& sof, const String& name = DEF_PARAM) const;

  // equality methods
  //
  bool8 eq(const Prediction& arg) const;

  // method: new
  //
  static void* operator new(size_t size) {
    return mgr_d.get();
  }

  // method: new[]
  //
  static void* operator new[](size_t size) {
    return mgr_d.getBlock(size);
  }

  // method: delete
  //
  static void operator delete(void* ptr) {
    mgr_d.release(ptr);
  }

  // method: delete[]
  //
  static void operator delete[](void* ptr) {
    mgr_d.releaseBlock(ptr);
  }
  
  // method: setGrowSize
  // 
  static bool8 setGrowSize(int32 grow_size) {
    return mgr_d.setGrow(grow_size);
  }

  // other memory-management methods
  //
  bool8 clear(Integral::CMODE ctype = Integral::DEF_CMODE);

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  set methods
  //
  //---------------------------------------------------------------------------
 
  // method: setAlgorithm
  //
  bool8 setAlgorithm(ALGORITHM algorithm) {
    algorithm_d = algorithm;
    is_valid_d = false;
    return true;  
  }
  
  // method: setImplementation
  //
  bool8 setImplementation(IMPLEMENTATION implementation) {
    implementation_d = implementation;
    is_valid_d = false;
    return true;  
  }
  
  // method: setOrder
  //
  bool8 setOrder(int32 order) {
    order_d = order;
    is_valid_d = false;
    return true;  
  }
  
  // method: setDynRange
  //
  bool8 setDynRange(float dyn_range) {
    dyn_range_d = dyn_range;
    is_valid_d = false;
    return true;  
  }

  // method: set
  //
  bool8 set(ALGORITHM algo = DEF_ALGORITHM,
	      IMPLEMENTATION impl = DEF_IMPLEMENTATION,
	      int32 order = DEF_ORDER, float dyn_range = DEF_DYN_RANGE) {
    algorithm_d = algo;
    implementation_d = impl;
    order_d = order;
    dyn_range_d = dyn_range;
    is_valid_d = false;
    return true;  
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  get methods
  //
  //---------------------------------------------------------------------------

  // method: getAlgorithm
  //
  ALGORITHM getAlgorithm() const {  
    return algorithm_d;
  }

  // method: getImplementation
  //
   IMPLEMENTATION getImplementation() const {  
    return implementation_d;
  }

  // method: getOrder
  //
  int32 getOrder() const {
    return order_d;
  }

  // method: getDynRange
  //
  float getDynRange() const {
    return dyn_range_d;
  }
  
  // method: get
  //
  bool8 get(ALGORITHM& algorithm, IMPLEMENTATION& implementation,
	      int32& order, float& dyn_range) const {
    algorithm = algorithm_d;
    implementation = implementation_d;
    order = order_d;
    dyn_range = dyn_range_d;
    return true;  
  }

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  computational methods
  //
  //---------------------------------------------------------------------------

  bool8 compute(VectorFloat& output, const VectorFloat& input,
		  AlgorithmData::COEF_TYPE input_coef_type = DEF_COEF_TYPE,
		  int32 index = DEF_CHANNEL_INDEX);
  bool8 compute(VectorFloat& output, const MatrixFloat& input,
		  AlgorithmData::COEF_TYPE input_coef_type = DEF_COEF_TYPE,
		  int32 index = DEF_CHANNEL_INDEX);
  bool8 compute(VectorFloat& output, float& err_energy,
		  const VectorFloat& input,
		  AlgorithmData::COEF_TYPE input_coef_type = DEF_COEF_TYPE,
		  int32 index = DEF_CHANNEL_INDEX);
  bool8 compute(VectorFloat& output, float& err_energy,
		  const MatrixFloat& input,
		  AlgorithmData::COEF_TYPE input_coef_type = DEF_COEF_TYPE,
		  int32 index = DEF_CHANNEL_INDEX);

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  public methods required by the AlgorithmBase interface contract
  //
  //---------------------------------------------------------------------------

  // assign method
  //
  bool8 assign(const AlgorithmBase& arg);

  // equality method
  //  
  bool8 eq(const AlgorithmBase& arg) const;
  
  // method: className
  //
  const String& className() const {
    return CLASS_NAME;
  }
  
  // computational methods
  //
  bool8 apply(Vector<AlgorithmData>& output,
		const Vector< CircularBuffer<AlgorithmData> >& input);
  
  // method to set the parser 
  //
  bool8 setParser(SofParser* parser);

  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:

  // algorithm-specific i/o methods
  //
  bool8 readDataCommon(Sof& sof, const String& pname,
                         int32 size = SofParser::FULL_OBJECT,
                         bool8 param = true, bool8 nested = false);
  bool8 writeDataCommon(Sof& sof, const String& pname) const;

  // algorithm and implementation specific computational methods
  //
  bool8 computeAutoDurbin(VectorFloat& output, float& err_energy,
			    const VectorFloat& input);
  bool8 computeCovarCholesky(VectorFloat& output, float& err_energy,
			       const MatrixFloat& input);
  bool8 computeLatticeBurg(VectorFloat& output, float& err_energy,
			     const VectorFloat& input);
  bool8 computeReflectionStepDown(VectorFloat& output,
				    const VectorFloat& input);
  bool8 computeLogAreaKellyLochbaum(VectorFloat& output,
				      const VectorFloat& input);  
};

// end of include file
// 
#endif
