// file: $isip_ifc/class/algo/Rps/rps_03.cc
// version: $Id: rps_03.cc 10484 2006-03-14 23:50:33Z srinivas $
//

// isip include files
//
#include "Rps.h"

//------------------------------------------------------------------
// these methods have to be in the same file so they can use the same
// static parser pointer
//------------------------------------------------------------------

// declare a static sof parser pointer
//
static SofParser* parser_d = (SofParser*)NULL;

// method: setParser
//
// arguments:
//  SofParser* parser: (input) sof file object
//
// return: a bool8 value indicating status
//
// this method sets the parser from the parent object to be used in
// the next readData method
//
bool8 Rps::setParser(SofParser* parser_a) {

  // set the parser
  //
  parser_d = parser_a;

  // exit gracefully
  //
  return true;
}

// method: read
//
// arguments:
//  Sof& sof: (input) sof file object
//  int32 tag: (input) sof object instance tag
//  const String& name: (input) sof object instance name
//
// return: a bool8 value indicating status
//
// this method has the object read itself from an Sof file according
// to the specified name and tag
//
bool8 Rps::read(Sof& sof_a, int32 tag_a, const String& name_a) {

  // get the instance of the object from the Sof file
  //
  if (!sof_a.find(name_a, tag_a)) {
    return false;
  }

  // read the actual data from the sof file
  //
  return readData(sof_a);
}

// method: readData
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  int32 size: (input) size in bytes of object
//  bool8 param: (input) is the parameter name in the file?
//  bool8 nested: (input) are we nested?
//
// return: a bool8 value indicating status
//
// this method has the object read itself from an Sof file. it assumes
// that the Sof file is already positioned correctly
//
bool8 Rps::readData(Sof& sof_a, const String& pname_a,
			     int32 size_a, bool8 param_a,
			     bool8 nested_a) {
  
  // we need a parser
  // 
  if (parser_d == (SofParser*)NULL) {
    
    parser_d = new SofParser;
    
    if (nested_a) {
      parser_d->setNest();
    }
    
    // load the parse
    //
    if (!parser_d->load(sof_a, size_a)) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
                           __FILE__, __LINE__, Error::WARNING);
    }           
  }

  // set the debug level of the parser
  //
  parser_d->setDebug(debug_level_d.getIndex());
  
  // get the mode
  //
  if (parser_d->isPresent(sof_a, PARAM_CMODE)) {
    if (!CMODE_MAP.readElementData((int32&)cmode_d, sof_a, PARAM_CMODE,
                                  parser_d->getEntry(sof_a, PARAM_CMODE))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
                           __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    cmode_d = DEF_CMODE;
  }
  setComputeMode(cmode_d);
  
  // get algorithm
  //
  if (parser_d->isPresent(sof_a, PARAM_ALGORITHM)) {
    if (!ALGO_MAP.readElementData((int32&)algorithm_d, sof_a, PARAM_ALGORITHM,
				  parser_d->getEntry(sof_a,
						     PARAM_ALGORITHM))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readData", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    algorithm_d = DEF_ALGORITHM;
  }

  // check known algorithms
  //
  if ((algorithm_d == TIME_DELAY_EMBEDDING) ||
      (algorithm_d == SVD_EMBEDDING)) {
    return readDataCommon(sof_a, pname_a, param_a, nested_a);
  }
    
  // exit ungracefully
  //
  else {
    return Error::handle(name(), L"readData",
			 ERR_UNKALG, __FILE__, __LINE__);
  }
}

// method: readDataCommon
//
// arguments:
//  Sof& sof: (input) sof file object
//  const String& pname: (input) parameter name
//  int32 size: (input) size in bytes of object (or FULL_OBJECT)
//  bool8 param: (input) is the parameter name in the file?
//  bool8 nested: (input) are we nested?
//
// return: a bool8 value indicating status
//
// this method assumes the algorithm has been read, and proceeds to read
// the rest of the object based on the value of the algorithm
//
bool8 Rps::readDataCommon(Sof& sof_a, const String& pname_a,
				   int32 size_a, bool8 param_a,
				   bool8 nested_a) {

  // get the implementation
  //
  if (parser_d->isPresent(sof_a, PARAM_IMPLEMENTATION)) {
    if (!IMPL_MAP.readElementData((int32&)implementation_d, sof_a,
				  PARAM_IMPLEMENTATION,
				  parser_d->getEntry(sof_a,
						     PARAM_IMPLEMENTATION))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    implementation_d = DEF_IMPLEMENTATION;
  }

  // get the embedding dimension
  //
  if (parser_d->isPresent(sof_a, PARAM_EMBED_DIMENSION)) {
    if (!embed_dimension_d.readData(sof_a, PARAM_EMBED_DIMENSION,
			  parser_d->getEntry(sof_a, PARAM_EMBED_DIMENSION))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    embed_dimension_d = DEF_EMBED_DIMENSION;
  }

  // get the delay for embedding
  //
  if (parser_d->isPresent(sof_a, PARAM_DELAY)) {
    if (!delay_d.readData(sof_a, PARAM_DELAY,
			  parser_d->getEntry(sof_a, PARAM_DELAY))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    delay_d = DEF_DELAY;
  }

  // get the SVD window size
  //
  if (parser_d->isPresent(sof_a, PARAM_SVD_WINDOW_SIZE)) {
    if (!svd_window_size_d.readData(sof_a, PARAM_SVD_WINDOW_SIZE,
			  parser_d->getEntry(sof_a, PARAM_SVD_WINDOW_SIZE))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    svd_window_size_d = DEF_SVD_WINDOW_SIZE;
  }
  
  // get threshold for SVD
  //
  if (parser_d->isPresent(sof_a, PARAM_SVD_THRESHOLD)) {
    if (!svd_threshold_d.readData(sof_a, PARAM_SVD_THRESHOLD,
			  parser_d->getEntry(sof_a, PARAM_SVD_THRESHOLD))) {

      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;

      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::READ,
			   __FILE__, __LINE__, Error::WARNING);
    }      
  }
  else {
    svd_threshold_d = DEF_SVD_THRESHOLD;
  }
  
  // get the debug level
  //
  if (parser_d->isPresent(sof_a, PARAM_DBGL)) {
    if (!debug_level_d.readData(sof_a, PARAM_DBGL,
				parser_d->getEntry(sof_a, PARAM_DBGL))) {
      
      // delete the parser
      //
      delete parser_d;
      parser_d = (SofParser*)NULL;
      
      // return a warning message
      //
      return Error::handle(name(), L"readDataCommon", Error::IO,
			   __FILE__, __LINE__, Error::WARNING);
    }
  }
  else {
    debug_level_d = Integral::DEF_DEBUG;
  }

  // check that all parameters are accounted for
  //
  if (!parser_d->checkParams(sof_a)) {

    // delete the parser
    //
    delete parser_d;
    parser_d = (SofParser*)NULL;

    // return a warning message
    //
    return Error::handle(name(), L"readDataCommon", Error::IO,
			 __FILE__, __LINE__, Error::WARNING);
  }

  // delete the parser
  //
  delete parser_d;
  parser_d = (SofParser*)NULL;
  
  // exit gracefully
  //
  return true;
}
