// file: $isip_ifc/class/sp/SignalDetector/sigd_09.cc
// version: $Id: sigd_09.cc 10210 2005-08-12 20:24:51Z stanley $
//

// isip include files
//
# include "SignalDetector.h"


// method: cbUttInProgress
//
// arguments: 
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 vlaue indicating whether we have a start of utterance
//
// a start of utterance is defined as the point at which we have a
// run of SIGN states. this does not mean this is the final start time
// of the utterance (which is found from some postprocessing). it simply
// means we are confident there is an utterance in progress.
//
bool8 SignalDetector::cbUttInProgress(int32 channel_index_a) {
  
  // make sure there the number of frames between the current frame
  // and the release frame is more than the utterance_minimum_duration_d
  //
  if((egy_cur_d(channel_index_a) - egy_rlse_d(channel_index_a)) < utterance_minimum_sample_duration_d) {
    return false;
  }
  
  // start from the egy_cur_d
  //
  int32 nf = egy_cur_d(channel_index_a);
  
  // try to find a run of signal states of duration utterance_minimum_duration_d
  // backwards from egy_cur_d
  //
  for(int32 counter = 0;counter<utterance_minimum_sample_duration_d; counter++) {
    
    if (nf == egy_rlse_d(channel_index_a)) {
      return Error::handle(name(), L"Error in circular buffer in cbUttInProgress function",Error::ARG,__FILE__, __LINE__);
    }
    
    if(egy_d(channel_index_a)(nf - egy_rlse_d(channel_index_a)) != SM_STATE_SIGN) {
      return false;
    }
    nf--;
  }
  
  // return gracefully
  //
  return true;
}

// method: cbUttNotInProgress
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating whether we have a noise state
//
// this method is normally called after an utterance has been in progress
// to determine when the end of utterance has occurred.
//
bool8 SignalDetector::cbUttNotInProgress(int32 channel_index_a) {
  
  // make sure that the number of frames between the egy_cur_d and
  // egy_rlse_d is greater than the minimum utterance seperation
  //
  if((egy_cur_d(channel_index_a) - egy_rlse_d(channel_index_a)) < utterance_minimum_sample_separation_d) {
    
    return false;
  }
  
  // start from the current frame
  //
  int32 nf = egy_cur_d(channel_index_a);
  
  // try to find a run of noise states equal to minimum utterance
  // separation backwards starting from the current frame
  //
  for(int32 counter = 0;counter<utterance_minimum_sample_separation_d; counter++) {
    
    if (nf == egy_rlse_d(channel_index_a)) {
      return Error::handle(name(), L"Error in circular buffer in cbUttNotInProgress function",Error::ARG,__FILE__, __LINE__);      
    }
    if(egy_d(channel_index_a)(nf - egy_rlse_d(channel_index_a)) != SM_STATE_NOIS) {
      return false;
    }
    nf--;
  }
  
  // return gracefully
  //
  return true;
  
}

// method: cbScanForPattern
//
// arguments:
//  Vector<Long> pattern: (input) pattern to be matched
//  int32 length: (input) length of the pattern
//  int32 index: (input) position from which to start scan 
//  int32 channel_index: (input) channel index 
//
// return: an int32 indicating the position of the match
//  a value less than zero indicates no match
//
// this method scans the circular buffer for the first occurrence of the
// pattern supplied as an argument.
//
int32 SignalDetector::cbScanForPattern(Vector<Long> pattern_a, int32 length_a , int32 index_a , int32 channel_index_a) {
  
  int32 pointer = index_a;
  
  int32 nf_stop = egy_rlse_d(channel_index_a);
  
  // make sure the number of frames between the release frame and the
  // "index" frame is more or equal to the length_a
  //
  for (int32 nf = 0; nf < length_a; nf++) {
    nf_stop++;
    if (nf_stop == pointer)
      {
	return (int32)-1;
      }
    
  }
  
  // start from the user given index and go towards the release frame
  // trying to find a match for the pattern
  //
  for (int32 nf = pointer ; nf >= nf_stop; nf--) {
    
    int32 num_matches = 0;
    int32 n = nf;
    for (int32 counter = 0; counter < length_a; counter++) {
      
      // match the frame state to that of the pattern
      //
      if(egy_d(channel_index_a)(n-egy_rlse_d(channel_index_a))== pattern_a (counter)) {
	num_matches++;
      }
      n--;
      
      if (n < egy_rlse_d(channel_index_a)) {
	return Error::handle(name(), L"Exception in the circular buffer in cbScanForPattern function",Error::ARG,__FILE__, __LINE__);
      }
      
    }
    //check to see if a pattern is found
    //
    if (num_matches == length_a) {
      
      return nf;
    }
  }
  return -1; 
}



// method: cbSmoothSignal
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating the status
//
// this method removes short bursts of noise that do not meet the
// duration constraints.
//
// this algorithm works as follows:
//  (1) checks from the current position if there are min_utt_dur consecutive
//      frames assigned to state SIGN
//  (2) scans backward to find another min_utt_dur consecutive frames
//      assigned to state SIGN previous to the frames in step (1)
//  (3) if the time difference between these two groups of frames
//      is less than min_utt_sep, assign them to SIGN
//  
bool8 SignalDetector::cbSmoothSignal(int32 channel_index_a) {
  
  Vector<Long> temp;
  int32 write_pointer = egy_cur_d(channel_index_a);
  
  // start from the current frame and go backward towards the release
  // frame and try to find a run of signla states that is less than
  // the minimum utterance duration
  //
  for (int32 counter = 0; counter <utterance_minimum_sample_duration_d; counter++) {
    
    if(egy_d(channel_index_a)(write_pointer - egy_rlse_d(channel_index_a)) != SM_STATE_SIGN) {
      return false;
    }
    write_pointer--;
    if (write_pointer == egy_rlse_d(channel_index_a)) {
      return false;
    }  
  }
  
  while (write_pointer != egy_rlse_d(channel_index_a)) {
    
    if (egy_d(channel_index_a)(write_pointer - egy_rlse_d(channel_index_a)) != SM_STATE_SIGN) {
      break;
    } 
    write_pointer--;
  }
  if (write_pointer == egy_rlse_d(channel_index_a)) {
    return false;
  }
  
  // try to find a match for sig_patn_d pattern starting from the
  // write_pointer index
  //
  int32 match_index = cbScanForPattern(sig_patn_d , sig_patn_len_d, write_pointer, channel_index_a);
  
  // if no match is found return false
  //
  if (match_index < 0) {
    return false;
  }
  
  // get the difference between the match index and the write pointer
  //
  int32 index_difference = abs(write_pointer - match_index) ;
  
  // make sure the difference is less than the minimum utterance
  // separation
  //
  if (index_difference < utterance_minimum_sample_separation_d) {
    
    // read the data into a temporary vector
    //
    int32 length = egy_d(channel_index_a).getNumForward();
    egy_d(channel_index_a).getElement(temp, (int32)(length + 1), 0);
    
    while (index_difference > 0) {
      
      // set the state to signal state
      //
      temp(write_pointer - egy_rlse_d(channel_index_a)) = SM_STATE_SIGN;
      index_difference--;
      write_pointer--;
    }
    
    // write back the vector into the circular buffer
    //
    CircularBuffer<Long> temp_egy_d(egy_len_d);
    for (int32 i = 0; i < egy_rlse_d(channel_index_a); i++) {
      temp_egy_d.append((Long)0);
    }
    temp_egy_d.append(temp);
    temp_egy_d.setRead(egy_rlse_d(channel_index_a));
    egy_d(channel_index_a).assign(temp_egy_d);
  }
  
  // return gracefully
  //
  return true;
  
}


// method: cbSmoothNoise
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating the status
//
// this method removes short bursts of energy that do not meet the
// duration constraints and occur within a run of noise.
//
// this algorithm works as follows:
//  (1) checks from the current position if there are min_utt_dur consecutive
//      frames assigned to state NOIS
//  (2) scans backward to find another min_utt_dur consecutive frames
//      assigned to state NOIS previous to the frames in step (1)
//  (3) if the time difference between these two groups of frames
//      is less than min_utt_dur, assign them to NOIS
//
bool8 SignalDetector::cbSmoothNoise(int32 channel_index_a) {
  
  
  Vector<Long> temp;
  int32 write_pointer = egy_cur_d(channel_index_a);
  
  // make sure there are min_utt_dur consecutive frames
  //
  for (int32 counter = 0; counter < utterance_minimum_sample_duration_d; counter++) {
    
    if (egy_d(channel_index_a)(write_pointer - egy_rlse_d(channel_index_a)) != SM_STATE_NOIS) {
      return false;
    }
    write_pointer--;
    if (write_pointer == egy_rlse_d(channel_index_a)) {
      return false;
    }  
  }
  
  while (write_pointer != egy_rlse_d(channel_index_a)) {
    
    if (egy_d(channel_index_a)(write_pointer - egy_rlse_d(channel_index_a)) != SM_STATE_NOIS) {
      break;
      
    } 
    write_pointer--;
  }
  if (write_pointer == egy_rlse_d(channel_index_a)) {
    return false;
  }
  
  // scan backward to find another minimum uterance duration of
  // consecutive frames
  //
  int32 match_index = cbScanForPattern(nse_patn_d , utterance_minimum_sample_duration_d, write_pointer, channel_index_a);
  
  if (match_index < 0) {
    return false;
  }
  
  // see if the difference between them is less than min_utt_dur
  //
  int32 index_difference = abs(write_pointer - match_index) ;
  
  // if the difference is less than min_utt_dur assign the states to
  // noise
  //
  if (index_difference < utterance_minimum_sample_duration_d) {
    
    int32 length = egy_d(channel_index_a).getNumForward();
    egy_d(channel_index_a).getElement(temp, (int32)(length + 1), 0);
    
    while (index_difference > 0) {
      
      // read the circular buffer into an temporary array
      //
      temp(write_pointer - egy_rlse_d(channel_index_a))= SM_STATE_NOIS;
      index_difference--;
      write_pointer--;
    }
    
    // write back the modified vector to the circular buffer
    //
    CircularBuffer<Long> temp_egy_d(egy_len_d);
    for (int32 i = 0; i < egy_rlse_d(channel_index_a); i++) {
      temp_egy_d.append((Long)0);
    }
    temp_egy_d.append(temp);
    temp_egy_d.setRead(egy_rlse_d(channel_index_a));
    egy_d(channel_index_a).assign(temp_egy_d);
  }
  return true;
  
}

// method: cbRemoveShortNoiseBursts
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating the status
//
// this method removes short bursts of noise that do not meet the
// duration constraints. note that this method starts from the
// back of the circular buffer (egy_rlse_d).
//
// this algorithm works as follows:
//  (1) finds a run of states that toggle between NOIS and NTOS
//  (2) assigns them to NOIS
//
bool8 SignalDetector::cbRemoveShortNoiseBursts(int32 channel_index_a) {
  
  Vector<Long> temp;
  
  int32 index = egy_cur_d(channel_index_a) -
    utterance_minimum_sample_separation_d;
  
  if ((index - egy_rlse_d(channel_index_a)) < 0) {
    return false;
  }
  
  // read the circular buffer into an temporary array
  //
  int32 length = egy_d(channel_index_a).getNumForward();
  egy_d(channel_index_a).getElement(temp, (int32)(length + 1), 0);  
  
  // assign the ntos states to noise state
  //
  while (index != egy_rlse_d(channel_index_a)) {
    if (temp(index - egy_rlse_d(channel_index_a)) == SM_STATE_NTOS) {
      temp(index - egy_rlse_d(channel_index_a)) = SM_STATE_NOIS;
    }
    index--;
  }
  
  // write the modified vector back to the circular buffer
  //
  CircularBuffer<Long> temp_egy_d(egy_len_d);
  for (int32 i = 0; i < egy_rlse_d(channel_index_a); i++) {
    temp_egy_d.append((Long)0);
  }
  temp_egy_d.append(temp);
  temp_egy_d.setRead(egy_rlse_d(channel_index_a));
  egy_d(channel_index_a).assign(temp_egy_d);  
  
  // return gracefully
  //
  return true;
}

// method: cbRemoveLongNoiseBursts
//
// arguments:
//  int32 channel_index: (input) audio channel index
// return: a bool8 indicating the status
//
// this method removes int32 bursts of noise. note that:
// 
//  - it works forward in time starting at egy_rlse_d
//  - it can only remove utterance_minimum_sample_separation_d
//    frames of noise when it finds twice this much
// 
bool8 SignalDetector::cbRemoveLongNoiseBursts(int32 channel_index_a) {
  
  // set the variable to twice as that utterance_minimum_separation_d
  //
  int32 num_states_req = 2 * utterance_minimum_sample_separation_d;
  int32 nf = egy_rlse_d(channel_index_a);  
  
  // try to find a run of noise states twice as that of the minimum
  // utterance separation
  //
  for (int32 counter = 0; counter < num_states_req; counter++) {
    
    if (egy_rlse_d(channel_index_a) == (Long) 0) {
      if (egy_d(channel_index_a)(nf) != SM_STATE_NOIS) {
	return false;
      } 
    }
    
    else {
      if (egy_d(channel_index_a)(nf - egy_rlse_d(channel_index_a))
	  != SM_STATE_NOIS) {
	return false;
      }
    }
    nf++;
    
    // id current frame is reached return false
    //
    if (nf == egy_cur_d(channel_index_a)) {
      return false;
    }
    
  }
  
  // if a run of noise states equal to twice that of the minimum
  // utterance separation is found move teh release pointer by a
  // utterance_minimum_sample_separation_d length
  //
  egy_d(channel_index_a).setRead(utterance_minimum_sample_separation_d);
  egy_rlse_d(channel_index_a) = egy_rlse_d(channel_index_a) + utterance_minimum_sample_separation_d;
  
  // return gracefully
  //
  return true;
}



// method: cbSmoothTransitions
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating the status
//
// this method turns short bursts of energy transitions that occur prior to
// min_utt_dur of SIGN states into either NOIS or SIGN.
//
// this algorithm works as follows:
//
//  (1) checks from the current position if there are min_utt_dur consecutive
//      frames assigned to state SIGN.  the time at which these SIGN state
//      first occur is called 't2'.
//  (2) scans backward to find another min_utt_dur consecutive frames
//      assigned to state NOIS previous to the frames in step (1).  the end of
//      these NOIS states is called 't1'.
//  (3) for all frames between t1 and t2, a sum is accumulated, with a SIGN
//      state adding 1, NOIS state adding 0, and transition states adding 1/2.
//  (4) if this sum is greater that the total number of NOIS states in this
//      interval, then all state in this interval are changed to SIGN, else
//      all states are changed to NOIS.
//
bool8 SignalDetector::cbSmoothTransitions(int32 channel_index_a) {
  
  Vector<Long> temp;
  int32 pointer = egy_cur_d(channel_index_a);
  int32 counter = 0;
  
  // check from the current position for a run of signal states of
  // minimum utterance duration
  //
  while ( counter < utterance_minimum_sample_duration_d) {
    if (egy_d(channel_index_a)(pointer - egy_rlse_d(channel_index_a))
	!= SM_STATE_SIGN) {
      return false;
    }
    counter++;
    pointer--;
    if (pointer < 0) {
      return false;
    }
    
  }
  int32 t2 = pointer;
  
  // scan backwards to another run of noise states
  //
  int32 t1 = cbScanForPattern(nse_patn_d, utterance_minimum_sample_separation_d, pointer, channel_index_a);
  if (t1 < 0) {
    return false;
  }
  
  // a good explanation for the logic below can be found above in this
  // functions header
  //
  float64 sum_sign = 0.0;
  float64 sum_nois = 0.0;
  for (counter = t1; counter < t2; counter++) {
    if (egy_d(channel_index_a)(counter - egy_rlse_d(channel_index_a)) == SM_STATE_SIGN){
      sum_sign += SM_WGT_SIGN;
    }
    else if (egy_d(channel_index_a)(counter - egy_rlse_d(channel_index_a)) == SM_STATE_NOIS) {
      sum_nois += SM_WGT_NOIS;
    }
    else {
      sum_sign += SM_WGT_TRAN;
    }
  }
  
  // make a decision on whether it is signal or noise
  //
  int32 new_val;
  
  if (sum_sign > sum_nois) {
    new_val = SM_STATE_SIGN;
  }
  else {
    new_val = SM_STATE_NOIS;
  }
  
  // set the frames to the new value
  //
  t1++;
  t2++;
  int32 length = egy_d(channel_index_a).getNumForward();
  egy_d(channel_index_a).getElement(temp, (int32)(length + 1), 0);  
  
  for (counter = t1; counter < t2; counter++) {
    temp(counter - egy_rlse_d(channel_index_a)) = new_val;
  }
  
  // write the temporary vector back to the circular buffer
  //
  CircularBuffer<Long> temp_egy_d(egy_len_d);
  for (int32 i = 0; i < egy_rlse_d(channel_index_a); i++) {
    temp_egy_d.append((Long)0);
  }
  temp_egy_d.append(temp);
  temp_egy_d.setRead(egy_rlse_d(channel_index_a));
  egy_d(channel_index_a).assign(temp_egy_d);
  
  // exit gracefully
  //
  return true;
}



// method: cbRemoveShortSignalBursts
//
// arguments:
//  int32 channel_index: (input) audio channel index
//
// return: a bool8 indicating the status
//
// this method removes short bursts of signal that do not meet the
// duration constraints.
//
// this algorithm works as follows:
//  (1) checks from the current position if there are utterance_minimum_separation_dur
//      consecutive frames assigned to state NOIS
//  (2) scans backward to find the next min_sep_dur consecutive frames
//      assigned to NOIS states
//  (3) checks if the time difference between these two groups of frames
//      is less than min_utt_sep, assigns them to NOIS
//
bool8 SignalDetector::cbRemoveShortSignalBursts(int32 channel_index_a) {
  
  // check from the current position for a run of noise states
  //
  int32 nf = egy_cur_d(channel_index_a);
  for (int32 counter = 0; counter < utterance_minimum_sample_separation_d; counter++) {
    if (egy_d(channel_index_a)(nf - egy_rlse_d(channel_index_a)) != SM_STATE_NOIS) {
      return false;
    }
    nf--;
    if (nf == egy_rlse_d(channel_index_a)) {
      return false;
    }
  }
  
  // check if the next state is non-NOIS
  //
  int32 start_index = nf - 1;
  if ((start_index == egy_rlse_d(channel_index_a)) ||
      (egy_d(channel_index_a)(start_index - egy_rlse_d(channel_index_a))
       == SM_STATE_NOIS)) {
    return false;
  }
  
  // find the next run of NOIS states
  //
  int32 end_index = start_index;
  int32 last_index = egy_rlse_d(channel_index_a) + utterance_minimum_sample_separation_d;
  
  while (end_index != last_index) {
    
    // count the number of frames that are assigned to the state NOIS
    //
    int32 num_states_nois = 0;
    nf = end_index;
    int32 counter = 0;
    
    while (counter < utterance_minimum_sample_separation_d) {
      if (egy_d(channel_index_a)(nf - egy_rlse_d(channel_index_a))
	  != SM_STATE_NOIS) {
	break;
      }
      else {
	num_states_nois++;
      }
      
      // update counters
      //
      counter++;
      nf--;
    }
    
    // check if we have a match
    //
    if (num_states_nois == utterance_minimum_sample_separation_d) {
      
      // check the duration of the burst
      //
      if (abs(start_index - end_index) < utterance_minimum_sample_duration_d) {
	
	// assign the states to NOIS
	//
	while (start_index != end_index) {
	  egy_d(channel_index_a)(start_index - egy_rlse_d(channel_index_a)) = SM_STATE_NOIS;
	  start_index--;
	}
	
	// exit
	//
	return true;
      }
    }
    
    // decrement counter
    //
    end_index--;
  }
  
  // exit gracefully - no bursts found
  //
  return false;
}
