// file: $isip/class/stat/SupportVectorModel/SupportVectorModel.h
// version: $Id: SupportVectorModel.h 9378 2003-12-23 20:56:11Z alphonso $
//

// make sure definitions are only made once
//
#ifndef ISIP_SUPPORT_VECTOR_MODEL
#define ISIP_SUPPORT_VECTOR_MODEL

// isip include files
//
#ifndef ISIP_SIGMOID
#include <Sigmoid.h>
#endif

#ifndef ISIP_VECTOR
#include <Vector.h>
#endif

#ifndef ISIP_KERNEL
#include <Kernel.h>
#endif

#ifndef ISIP_VECTOR_FLOAT
#include <VectorFloat.h>
#endif

#ifndef ISIP_MEMORY_MANAGER
#include <MemoryManager.h>
#endif

#ifndef ISIP_STATISTICAL_MODEL_BASE
#include <StatisticalModelBase.h>
#endif

// SupportVectorModel: a class to compute distance of the test vector
// from a hyperplane in transformed space
//
//                     N
//                   ----                      
// distance = bias + \    [ alpha(i) * kernel (x, sv[i]) ]
//                   /___                    i
//                     i
//
//  'sv' is the support vector
//  'N' is the number of support vectors
//  'alpha(i)' is the weigth for i-th support vector
//  'bias' is the distance of hyperplane from the origin
//
// resulting distance (positive or negative) can be transformed to
// posterior probability using e.g. sigmoid transformation
//
// probability = sigmoid (distance)
//
class SupportVectorModel : public StatisticalModelBase {

  //---------------------------------------------------------------------------
  //
  // public constants
  //
  //---------------------------------------------------------------------------
public:

  // define the class name
  //
  static const String CLASS_NAME;
  
  //----------------------------------------
  //
  // other important constants
  //
  //----------------------------------------
  
  enum ALGORITHM { SINGLE_KERNEL = 0, MULTIPLE_KERNEL,
                   DEF_ALGORITHM = SINGLE_KERNEL };

  enum IMPLEMENTATION { SIGMOID = 0, LOGISTIC, LOGSYM, TANH, SINUS, GAUSS,
			RANDOM, DEF_IMPLEMENTATION = SIGMOID };

  static const float64 MAX_LOG_DURATION = 3.912;

  // define static NameMap objects for the enumerated values
  //
  static const NameMap ALGO_MAP;
  static const NameMap IMPL_MAP;

  //----------------------------------------
  //
  // i/o related constants
  //
  //----------------------------------------  
  
  static const String DEF_PARAM;
  static const String PARAM_ALGORITHM;
  static const String PARAM_IMPLEMENTATION;
  static const String PARAM_SUPPORT_VECTORS;
  static const String PARAM_KERNELS;
  static const String PARAM_ALPHAS;
  static const String PARAM_BIAS;
  static const String PARAM_SIGMOID;
  static const String DEF_COMMENT_TAG;

  //----------------------------------------
  //
  // error codes
  //
  //----------------------------------------  

  static const int32 ERR = 60350;
    
  //---------------------------------------------------------------------------
  //
  // protected data
  //
  //---------------------------------------------------------------------------
protected:

  // define algorithm and implementation
  //
  ALGORITHM algorithm_d;
  IMPLEMENTATION implementation_d;

  // support vectors, their kernels and alphas
  //
  Vector<VectorFloat> support_vectors_d;
  Vector<Kernel> kernels_d;
  Vector<Double> alphas_d;
  Double bias_d;

  // define distance->probability transformation function
  // currently sigmoid
  //
  Sigmoid sigmoid_d;

  // memory manager
  //
  static MemoryManager mgr_d;
  
  //---------------------------------------------------------------------------
  //
  // required public methods
  //
  //---------------------------------------------------------------------------
public:

  // method: name
  //
  static const String& name() {
    return CLASS_NAME;
  }

  // other static methods
  //
  static bool8 diagnose(Integral::DEBUG debug_level);

  // debug methods
  //  setDebug is inherited from base class
  //
  bool8 debug(const unichar* msg) const;
  
  // method: destructor
  //
  ~SupportVectorModel() {}

  // method: default constructor
  //
  SupportVectorModel() {
    algorithm_d = DEF_ALGORITHM;
    implementation_d = DEF_IMPLEMENTATION;
    bias_d = 0;
  }

  // method: copy constructor
  //
  SupportVectorModel(const SupportVectorModel& arg) {
    assign(arg);
  }

  // assign methods
  //
  bool8 assign(const SupportVectorModel& arg);

  // method: operator=
  //
  SupportVectorModel& operator=(const SupportVectorModel& arg) {
    assign(arg);
    return *this;
  }
  
  // sofSize method
  //
  int32 sofSize() const;
  
  // method: sofAccumulatorSize
  //
  int32 sofAccumulatorSize() const {
    return Error::handle(name(), L"sofAccumulatorSize", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: sofOccupanciesSize
  //
  int32 sofOccupanciesSize() const {
    return Error::handle(name(), L"sofOccupanciesSize", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }    

  // other i/o methods
  //
  bool8 read(Sof& sof, int32 tag, const String& name = CLASS_NAME);
  bool8 write(Sof& sof, int32 tag, const String& name = CLASS_NAME) const;

  bool8 readData(Sof& sof, const String& pname = DEF_PARAM,
                   int32 size = SofParser::FULL_OBJECT,
                   bool8 param = true,
                   bool8 nested = false);
  bool8 writeData(Sof& sof, const String& pname = DEF_PARAM) const;

  // method: readAccumulator
  //
  bool8 readAccumulator(Sof& sof, int32 tag,
		    const String& cname = CLASS_NAME) {
    return Error::handle(name(), L"readAccumulator", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: writeAccumulator
  //
  bool8 writeAccumulator(Sof& sof, int32 tag,
		     const String& cname = CLASS_NAME) const {
    return Error::handle(name(), L"writeAccumulator", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: readAccumulatorData
  //
  bool8 readAccumulatorData(Sof& sof, const String& pname = DEF_PARAM,
			int32 size = SofParser::FULL_OBJECT,
			bool8 param = true,
			bool8 nested = false) {
    return Error::handle(name(), L"readAccumulatorData", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: writeAccumulatorData
  //
  bool8 writeAccumulatorData(Sof& sof,
			       const String& pname = DEF_PARAM) const {
    return Error::handle(name(), L"writeAccumulatorData", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }
  
  // method: readOccupancies
  //
  bool8 readOccupancies(Sof& sof, int32 tag,
			  const String& cname = CLASS_NAME) {
    return Error::handle(name(), L"readOccupancies", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: writeOccupancies
  //
  bool8 writeOccupancies(Sof& sof, int32 tag,
			   const String& cname = CLASS_NAME) const {
    return Error::handle(name(), L"writeOccupancies", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }
  
  // method: readOccupanciesData
  //
  bool8 readOccupanciesData(Sof& sof, const String& pname = DEF_PARAM,
			      int32 size = SofParser::FULL_OBJECT,
			      bool8 param = true,
			      bool8 nested = false) {
    return Error::handle(name(), L"readOccupanciesData", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: writeOccupanciesData
  //
  bool8 writeOccupanciesData(Sof& sof,
			       const String& pname = DEF_PARAM) const {
    return Error::handle(name(), L"writeOccupanciesData", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }
  
  // equality methods
  //
  bool8 eq(const SupportVectorModel& arg) const;

  // method: operator new
  //
  static void* operator new(size_t size) {
    return mgr_d.get();
  }

  // method: operator new[]
  //
  static void* operator new[](size_t size) {
    return mgr_d.getBlock(size);
  }

  // method: operator delete
  //
  static void operator delete(void* ptr) {
    mgr_d.release(ptr);
  }

  // method: operator delete[]
  //
  static void operator delete[](void* ptr) {
    mgr_d.releaseBlock(ptr);
  }

  // method: setGrowSize
  //
  static bool8 setGrowSize(int32 grow_size) {
    return mgr_d.setGrow(grow_size);
  }

  // method: clear
  //
  bool8 clear(Integral::CMODE cmode = Integral::DEF_CMODE) {
    support_vectors_d.clear(cmode);
    kernels_d.clear(cmode);
    alphas_d.clear(cmode);
    sigmoid_d.clear(cmode);
    return true;
  }
  
  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  additional methods unique to this class
  //
  //--------------------------------------------------------------------------

  // method: setAlgorithm
  // to set single or muti-kernel mode
  //
  bool8 setAlgorithm(ALGORITHM algo) {
    algorithm_d = algo;
    return true;
  }

  // method: getAlgorithm
  // to get what mode we are using (single/multi-kernel)
  //
  ALGORITHM getAlgorithm() {
    return algorithm_d;
  }

  // method: setImplementation
  // methods to set the distance->probability transformation
  //
  bool8 setImplementation(IMPLEMENTATION impl) {
    implementation_d = impl;
    return true;
  }

  // method: getImplementation
  // to get what distance->probability transformation we use
  //
  IMPLEMENTATION getImplemenation() {
    return implementation_d;
  }

  // methods to manipulate the support vectors, kernels and weights
  //
  bool8 setNumSupportVectors(int32 number,
			       ALGORITHM algorithm = DEF_ALGORITHM);

  // method: getSupportVector
  //
  VectorFloat& getSupportVector(int32 index) {
    return support_vectors_d(index);
  }

  // method: getSupportVectors
  //
  Vector<VectorFloat>& getSupportVectors() {
    return support_vectors_d;
  }  

  // method: getKernel
  //
  Kernel& getKernel(int32 index) {
    if (algorithm_d == SINGLE_KERNEL) {
      return kernels_d(index);
    }
    else {
      return kernels_d(0);
    }
  }

  // method: getKernels
  //
  Vector<Kernel>& getKernels() {
    return kernels_d;
  }  

  // method: getAlpha
  //
  Double& getAlpha(int32 index) {
    return alphas_d(index);
  }

  // method: getAlphas
  //
  Vector<Double>& getAlphas() {
    return alphas_d;
  }
  
  // method: getBias
  //
  Double& getBias() {
    return bias_d;
  }

  // method: getSigmoid
  //
  Sigmoid& getSigmoid() {
    if (!implementation_d == SIGMOID) {
      Error::handle(name(), L"getSigmoid - no sigmoid",
			   Error::ARG, __FILE__, __LINE__);
    }
    return sigmoid_d;
  }

  // get the SVM distance (this method will check dimensions first)
  //
  float64 getDistance(const VectorFloat& data);

  static bool8 generateSvmFeatures(Vector<Vector<VectorFloat> >& segm_feat,
				     const Vector<VectorFloat>& frame_feat);

  static bool8 generateSvmSegment(VectorFloat& segm_feat,
				    int32 feat_dim,
				    int32 start_frame, int32 end_frame,
				    bool8 duration_flag,
				    const Vector<VectorFloat>& frame_feat);

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  required for the base class interface contract
  //
  //--------------------------------------------------------------------------

  // StatisticalModelBase required methods
  //
  bool8 assign(const StatisticalModelBase& arg);
  bool8 eq(const StatisticalModelBase& arg) const;

  // set methods
  //
  bool8 setMode(MODE arg);

  // method: className
  //
  const String& className() const {
    return CLASS_NAME;
  }

  // initialization methods
  //
  bool8 init();

  // method: getLikelihood
  //
  float32 getLikelihood(const VectorFloat& input) {
    return Integral::exp(getLogLikelihood(input));
  }

  // computational methods
  //
  float32 getLogLikelihood(const VectorFloat& input);

  //---------------------------------------------------------------------------
  //
  // class-specific public methods:
  //  accumulate and update methods needed for training models
  //
  //---------------------------------------------------------------------------

  // method: getMean
  //
  bool8 getMean(VectorFloat& mean) {
    return Error::handle(name(), L"getMean", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: getCovariance
  //
  bool8 getCovariance(MatrixFloat& cov) {
    return Error::handle(name(), L"getCovariance", Error::NOT_IMPLEM,
			 __FILE__, __LINE__);
  }

  // method: resetAccumulators
  //
  bool8 resetAccumulators() {
    return Error::handle(name(), L"resetAccumulators",
			 Error::NOT_IMPLEM, __FILE__, __LINE__);
  }

  // method: getOccupancy
  //
  float64 getOccupancy() {
    return Error::handle(name(), L"getOccupancy",
			 Error::NOT_IMPLEM, __FILE__, __LINE__);
  }

  // method: setOccupancy
  //
  bool8 setOccupancy(float64 arg) {
    return Error::handle(name(), L"setOccupancy",
			 Error::NOT_IMPLEM, __FILE__, __LINE__);
  }  

  // method: getAccessCount
  //
  int32 getAccessCount() {
    return Error::handle(name(), L"getAccessCount",
			 Error::NOT_IMPLEM, __FILE__, __LINE__);
  }

  // method: setAccessCount
  //
  bool8 setAccessCount(int32 arg) {
    return Error::handle(name(), L"setAccessCount",
			 Error::NOT_IMPLEM, __FILE__, __LINE__);
  }  
  
  // method that accumulates the statistics for the model which are
  // needed to update the model parameters during training
  //
  bool8 accumulate(VectorDouble& param,
		     VectorFloat& data, bool8 precomp);

  // method that updates the statistical model parameters using the
  // accumulated statistics during training
  //  
  bool8 update(VectorFloat& varfloor, int32 min_count);

  // methods that initializes the statistical model parameters using
  // accumulated feature vectors
  //
  bool8 accumulate(VectorFloat& data);    
  bool8 initialize(VectorFloat& param);

  bool8 loadSvmLight(const String& input_filename);
  
  //---------------------------------------------------------------------------
  //
  // private methods
  //
  //---------------------------------------------------------------------------
private:
};

// end of include file
// 
#endif
